from django.db.models import Count, OuterRef, Subquery, IntegerField
from django.conf import settings
from django.utils.translation import ugettext_lazy as _
from rest_framework import decorators, response, status, exceptions, serializers as rf_serializers, generics

from waldur_core.core import exceptions as core_exceptions, validators as core_validators
from waldur_core.core import utils as core_utils
from waldur_core.structure import models as structure_models
from waldur_core.structure import signals as structure_signals
from waldur_core.structure import views as structure_views, filters as structure_filters
from waldur_core.structure import permissions as structure_permissions
from waldur_openstack.openstack import models as openstack_models
from waldur_openstack.openstack.apps import OpenStackConfig
from waldur_openstack.openstack_base.backend import OpenStackBackendError

from . import models, serializers, filters, executors


class TelemetryMixin(object):
    """
    This mixin adds /meters endpoint to the resource.

    List of available resource meters must be specified in separate JSON file in meters folder. In addition,
    mapping between resource model and meters file path must be specified
    in "_get_meters_file_name" method in "backend.py" file.
    """

    telemetry_serializers = {
        'meter_samples': serializers.MeterSampleSerializer
    }

    @decorators.detail_route(methods=['get'])
    def meters(self, request, uuid=None):
        """
        To list available meters for the resource, make **GET** request to
        */api/<resource_type>/<uuid>/meters/*.
        """
        resource = self.get_object()
        backend = resource.get_backend()

        meters = backend.list_meters(resource)

        page = self.paginate_queryset(meters)
        if page is not None:
            return self.get_paginated_response(page)

        return response.Response(meters)

    @decorators.detail_route(methods=['get'], url_path='meter-samples/(?P<name>[a-z0-9_.]+)')
    def meter_samples(self, request, name, uuid=None):
        """
        To get resource meter samples make **GET** request to */api/<resource_type>/<uuid>/meter-samples/<meter_name>/*.
        Note that *<meter_name>* must be from meters list.

        In order to get a list of samples for the specific period of time, *start* timestamp and *end* timestamp query
        parameters can be provided:

            - start - timestamp (default: one hour ago)
            - end - timestamp (default: current datetime)

        Example of a valid request:

        .. code-block:: http

            GET /api/openstack-instances/1143357799fc4cb99636c767136bef86/meter-samples/memory/?start=1470009600&end=1470843282
            Content-Type: application/json
            Accept: application/json
            Authorization: Token c84d653b9ec92c6cbac41c706593e66f567a7fa4
            Host: example.com
        """
        resource = self.get_object()
        backend = resource.get_backend()

        meters = backend.list_meters(resource)
        names = [meter['name'] for meter in meters]
        if name not in names:
            raise exceptions.ValidationError(_('Meter must be from meters list.'))
        if not resource.backend_id:
            raise exceptions.ValidationError(_('%s must have backend_id.') % resource.__class__.__name__)

        serializer = serializers.MeterTimestampIntervalSerializer(data=request.query_params)
        serializer.is_valid(raise_exception=True)
        start = serializer.validated_data['start']
        end = serializer.validated_data['end']

        samples = backend.get_meter_samples(resource, name, start=start, end=end)
        serializer = self.get_serializer(samples, many=True)

        return response.Response(serializer.data)

    def get_serializer_class(self):
        serializer = self.telemetry_serializers.get(self.action)
        return serializer or super(TelemetryMixin, self).get_serializer_class()


class OpenStackServiceViewSet(structure_views.BaseServiceViewSet):
    queryset = models.OpenStackTenantService.objects.all()
    serializer_class = serializers.ServiceSerializer


class OpenStackServiceProjectLinkViewSet(structure_views.BaseServiceProjectLinkViewSet):
    queryset = models.OpenStackTenantServiceProjectLink.objects.all()
    serializer_class = serializers.ServiceProjectLinkSerializer
    filter_class = filters.OpenStackTenantServiceProjectLinkFilter


class UsageReporter(object):
    """
    This class implements service for counting number of instances grouped
    by image and flavor name and by instance runtime status.
    Please note that even when flavors have different UUIDs they are treated
    as the same as long as they have the same name.
    This is needed because in OpenStack UUID is not stable for images and flavors.
    """
    def __init__(self, view, request):
        self.view = view
        self.request = request
        self.query = None

    def get_report(self):
        if self.request.query_params:
            self.query = self.parse_query(self.request)

        running_stats = self.get_stats(models.Instance.RuntimeStates.ACTIVE)
        created_stats = self.get_stats()
        qs = self.get_initial_queryset().values_list('name', flat=True).distinct()

        page = self.view.paginate_queryset(qs)
        result = self.serialize_result(page, running_stats, created_stats)
        return self.view.get_paginated_response(result)

    def serialize_result(self, queryset, running_stats, created_stats):
        result = []
        for name in queryset:
            result.append({
                'name': name,
                'running_instances_count': running_stats.get(name, 0),
                'created_instances_count': created_stats.get(name, 0),
            })
        return result

    def apply_filters(self, qs):
        if self.query:
            filter_dict = dict()
            if self.query.get('shared', None):
                filter_dict['service_project_link__service__settings__shared'] = self.query['shared']
            if self.query.get('service_provider', None):
                filter_dict['service_project_link__service__settings__uuid__in'] = self.query['service_provider']
                filter_dict['service_project_link__service__settings__type'] = 'OpenStackTenant'
            return qs.filter(**filter_dict)
        return qs

    def parse_query(self, request):
        serializer_class = serializers.UsageStatsSerializer
        serializer = serializer_class(data=request.query_params)
        serializer.is_valid(raise_exception=True)
        query = serializer.validated_data
        return query

    def get_initial_queryset(self):
        raise NotImplementedError

    def get_stats(self, runtime_state=None):
        raise NotImplementedError


class ImageUsageReporter(UsageReporter):

    def get_initial_queryset(self):
        return models.Image.objects.all()

    def get_stats(self, runtime_state=None):
        volumes = models.Volume.objects.filter(bootable=True)
        if runtime_state:
            volumes = volumes.filter(instance__runtime_state=runtime_state)
        rows = self.apply_filters(volumes).values('image_name').annotate(count=Count('image_name'))
        return {row['image_name']: row['count'] for row in rows}


class FlavorUsageReporter(UsageReporter):

    def get_initial_queryset(self):
        return models.Flavor.objects.all()

    def get_stats(self, runtime_state=None):
        instances = models.Instance.objects.all()
        if runtime_state:
            instances = instances.filter(runtime_state=runtime_state)
        rows = self.apply_filters(instances)\
            .values('flavor_name').annotate(count=Count('flavor_name'))
        return {row['flavor_name']: row['count'] for row in rows}


class ImageViewSet(structure_views.BaseServicePropertyViewSet):
    queryset = models.Image.objects.all().order_by('settings', 'name')
    serializer_class = serializers.ImageSerializer
    lookup_field = 'uuid'
    filter_class = filters.ImageFilter

    @decorators.list_route()
    def usage_stats(self, request):
        return ImageUsageReporter(self, request).get_report()


class FlavorViewSet(structure_views.BaseServicePropertyViewSet):
    """
    VM instance flavor is a pre-defined set of virtual hardware parameters that the instance will use:
    CPU, memory, disk size etc. VM instance flavor is not to be confused with VM template -- flavor is a set of virtual
    hardware parameters whereas template is a definition of a system to be installed on this instance.
    """
    queryset = models.Flavor.objects.all().order_by('settings', 'cores', 'ram', 'disk')
    serializer_class = serializers.FlavorSerializer
    lookup_field = 'uuid'
    filter_class = filters.FlavorFilter

    @decorators.list_route()
    def usage_stats(self, request):
        return FlavorUsageReporter(self, request).get_report()


class NetworkViewSet(structure_views.BaseServicePropertyViewSet):
    queryset = models.Network.objects.all().order_by('settings', 'type', 'is_external')
    serializer_class = serializers.NetworkSerializer
    lookup_field = 'uuid'
    filter_class = filters.NetworkFilter


class SubNetViewSet(structure_views.BaseServicePropertyViewSet):
    queryset = models.SubNet.objects.all().order_by('settings')
    serializer_class = serializers.SubNetSerializer
    lookup_field = 'uuid'
    filter_class = filters.SubNetFilter


class FloatingIPViewSet(structure_views.BaseServicePropertyViewSet):
    queryset = models.FloatingIP.objects.all().order_by('settings', 'address')
    serializer_class = serializers.FloatingIPSerializer
    lookup_field = 'uuid'
    filter_class = filters.FloatingIPFilter


class SecurityGroupViewSet(structure_views.BaseServicePropertyViewSet):
    queryset = models.SecurityGroup.objects.all().order_by('settings', 'name')
    serializer_class = serializers.SecurityGroupSerializer
    lookup_field = 'uuid'
    filter_class = filters.SecurityGroupFilter


class VolumeViewSet(structure_views.ImportableResourceViewSet):
    queryset = models.Volume.objects.all()
    serializer_class = serializers.VolumeSerializer
    filter_class = filters.VolumeFilter

    create_executor = executors.VolumeCreateExecutor
    update_executor = executors.VolumeUpdateExecutor
    pull_executor = executors.VolumePullExecutor

    def _volume_snapshots_exist(volume):
        if volume.snapshots.exists():
            raise core_exceptions.IncorrectStateException(_('Volume has dependent snapshots.'))

    delete_executor = executors.VolumeDeleteExecutor
    destroy_validators = [
        _volume_snapshots_exist,
        core_validators.StateValidator(models.Volume.States.OK, models.Volume.States.ERRED),
        core_validators.RuntimeStateValidator('available', 'error', 'error_restoring', 'error_extending', ''),
    ]

    def _is_volume_bootable(volume):
        if volume.bootable:
            raise core_exceptions.IncorrectStateException(_('Volume cannot be bootable.'))

    def _is_volume_instance_shutoff(volume):
        if volume.instance and volume.instance.runtime_state != models.Instance.RuntimeStates.SHUTOFF:
            raise core_exceptions.IncorrectStateException(_('Volume instance should be in shutoff state.'))

    def _is_volume_instance_ok(volume):
        if volume.instance and volume.instance.state != models.Instance.States.OK:
            raise core_exceptions.IncorrectStateException(_('Volume instance should be in OK state.'))

    @decorators.detail_route(methods=['post'])
    def extend(self, request, uuid=None):
        """ Increase volume size """
        volume = self.get_object()
        old_size = volume.size
        serializer = self.get_serializer(volume, data=request.data)
        serializer.is_valid(raise_exception=True)
        serializer.save()

        volume.refresh_from_db()
        executors.VolumeExtendExecutor().execute(volume, old_size=old_size, new_size=volume.size)

        return response.Response({'status': _('extend was scheduled')}, status=status.HTTP_202_ACCEPTED)

    extend_validators = [_is_volume_bootable,
                         _is_volume_instance_ok,
                         _is_volume_instance_shutoff,
                         core_validators.StateValidator(models.Volume.States.OK)]
    extend_serializer_class = serializers.VolumeExtendSerializer

    @decorators.detail_route(methods=['post'])
    def snapshot(self, request, uuid=None):
        """ Create snapshot from volume """
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        snapshot = serializer.save()

        executors.SnapshotCreateExecutor().execute(snapshot)
        return response.Response(serializer.data, status=status.HTTP_201_CREATED)

    snapshot_serializer_class = serializers.SnapshotSerializer

    @decorators.detail_route(methods=['post'])
    def create_snapshot_schedule(self, request, uuid=None):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        serializer.save()
        return response.Response(serializer.data, status=status.HTTP_201_CREATED)

    create_snapshot_schedule_validators = [core_validators.StateValidator(models.Volume.States.OK)]
    create_snapshot_schedule_serializer_class = serializers.SnapshotScheduleSerializer

    @decorators.detail_route(methods=['post'])
    def attach(self, request, uuid=None):
        """ Attach volume to instance """
        volume = self.get_object()
        serializer = self.get_serializer(volume, data=request.data)
        serializer.is_valid(raise_exception=True)
        serializer.save()

        executors.VolumeAttachExecutor().execute(volume)
        return response.Response({'status': _('attach was scheduled')}, status=status.HTTP_202_ACCEPTED)

    attach_validators = [core_validators.RuntimeStateValidator('available'),
                         core_validators.StateValidator(models.Volume.States.OK)]
    attach_serializer_class = serializers.VolumeAttachSerializer

    @decorators.detail_route(methods=['post'])
    def detach(self, request, uuid=None):
        """ Detach instance from volume """
        volume = self.get_object()
        executors.VolumeDetachExecutor().execute(volume)
        return response.Response({'status': _('detach was scheduled')}, status=status.HTTP_202_ACCEPTED)

    detach_validators = [_is_volume_bootable,
                         core_validators.RuntimeStateValidator('in-use'),
                         core_validators.StateValidator(models.Volume.States.OK)]

    importable_resources_backend_method = 'get_volumes_for_import'
    importable_resources_serializer_class = serializers.VolumeImportableSerializer
    import_resource_serializer_class = serializers.VolumeImportSerializer


class SnapshotViewSet(structure_views.ImportableResourceViewSet):
    queryset = models.Snapshot.objects.all().order_by('name')
    serializer_class = serializers.SnapshotSerializer
    update_executor = executors.SnapshotUpdateExecutor
    delete_executor = executors.SnapshotDeleteExecutor
    pull_executor = executors.SnapshotPullExecutor
    filter_class = filters.SnapshotFilter
    disabled_actions = ['create']

    @decorators.detail_route(methods=['post'])
    def restore(self, request, uuid=None):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        restoration = serializer.save()

        executors.SnapshotRestorationExecutor().execute(restoration)
        serialized_volume = serializers.VolumeSerializer(restoration.volume, context={'request': self.request})
        return response.Response(serialized_volume.data, status=status.HTTP_201_CREATED)

    restore_serializer_class = serializers.SnapshotRestorationSerializer
    restore_validators = [core_validators.StateValidator(models.Snapshot.States.OK)]

    @decorators.detail_route(methods=['get'])
    def restorations(self, request, uuid=None):
        snapshot = self.get_object()
        serializer = self.get_serializer(snapshot.restorations.all(), many=True)
        return response.Response(serializer.data, status=status.HTTP_200_OK)

    restorations_serializer_class = serializers.SnapshotRestorationSerializer

    importable_resources_backend_method = 'get_snapshots_for_import'
    importable_resources_serializer_class = serializers.SnapshotImportableSerializer
    import_resource_serializer_class = serializers.SnapshotImportSerializer


class InstanceAvailabilityZoneViewSet(structure_views.BaseServicePropertyViewSet):
    queryset = models.InstanceAvailabilityZone.objects.all().order_by('settings', 'name')
    serializer_class = serializers.InstanceAvailabilityZoneSerializer
    lookup_field = 'uuid'
    filter_class = filters.InstanceAvailabilityZoneFilter


class InstanceViewSet(structure_views.ImportableResourceViewSet):
    """
    OpenStack instance permissions
    ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

    - Staff members can list all available VM instances in any service.
    - Customer owners can list all VM instances in all the services that belong to any of the customers they own.
    - Project administrators can list all VM instances, create new instances and start/stop/restart instances in all the
      services that are connected to any of the projects they are administrators in.
    - Project managers can list all VM instances in all the services that are connected to any of the projects they are
      managers in.
    """
    queryset = models.Instance.objects.all()
    serializer_class = serializers.InstanceSerializer
    filter_class = filters.InstanceFilter
    filter_backends = structure_views.ResourceViewSet.filter_backends + (
        structure_filters.StartTimeFilter,
    )
    pull_executor = executors.InstancePullExecutor
    pull_serializer_class = rf_serializers.Serializer

    update_executor = executors.InstanceUpdateExecutor
    update_validators = partial_update_validators = [core_validators.StateValidator(models.Instance.States.OK)]

    def perform_create(self, serializer):
        instance = serializer.save()
        executors.InstanceCreateExecutor.execute(
            instance,
            ssh_key=serializer.validated_data.get('ssh_public_key'),
            flavor=serializer.validated_data['flavor'],
            is_heavy_task=True,
        )

    def _has_backups(instance):
        if instance.backups.exists():
            raise core_exceptions.IncorrectStateException(_('Cannot delete instance that has backups.'))

    def _can_destroy_instance(instance):
        if instance.state == models.Instance.States.ERRED:
            return
        if (instance.state == models.Instance.States.OK and
                instance.runtime_state == models.Instance.RuntimeStates.SHUTOFF):
            return
        if (instance.state == models.Instance.States.OK and
                instance.runtime_state == models.Instance.RuntimeStates.ACTIVE):
            raise core_exceptions.IncorrectStateException(_('Please stop the instance before its removal.'))
        raise core_exceptions.IncorrectStateException(_('Instance should be shutoff and OK or erred. '
                                                        'Please contact support.'))

    def destroy(self, request, uuid=None):
        """
        Deletion of an instance is done through sending a **DELETE** request to the instance URI.
        Valid request example (token is user specific):

        .. code-block:: http

            DELETE /api/openstacktenant-instances/abceed63b8e844afacd63daeac855474/ HTTP/1.1
            Authorization: Token c84d653b9ec92c6cbac41c706593e66f567a7fa4
            Host: example.com

        Only stopped instances or instances in ERRED state can be deleted.

        By default when instance is destroyed, all data volumes
        attached to it are destroyed too. In order to preserve data
        volumes use query parameter ?delete_volumes=false
        In this case data volumes are detached from the instance and
        then instance is destroyed. Note that system volume is deleted anyway.
        For example:

        .. code-block:: http

            DELETE /api/openstacktenant-instances/abceed63b8e844afacd63daeac855474/?delete_volumes=false HTTP/1.1
            Authorization: Token c84d653b9ec92c6cbac41c706593e66f567a7fa4
            Host: example.com

        """
        serializer = self.get_serializer(data=request.query_params, instance=self.get_object())
        serializer.is_valid(raise_exception=True)
        delete_volumes = serializer.validated_data['delete_volumes']
        release_floating_ips = serializer.validated_data['release_floating_ips']

        resource = self.get_object()
        force = resource.state == models.Instance.States.ERRED
        executors.InstanceDeleteExecutor.execute(
            resource,
            force=force,
            delete_volumes=delete_volumes,
            release_floating_ips=release_floating_ips,
            async=self.async_executor,
        )

        return response.Response({'status': _('destroy was scheduled')}, status=status.HTTP_202_ACCEPTED)

    destroy_validators = [_can_destroy_instance, _has_backups]
    destroy_serializer_class = serializers.InstanceDeleteSerializer

    @decorators.detail_route(methods=['post'])
    def change_flavor(self, request, uuid=None):
        instance = self.get_object()
        old_flavor_name = instance.flavor_name
        serializer = self.get_serializer(instance, data=request.data)
        serializer.is_valid(raise_exception=True)
        serializer.save()

        flavor = serializer.validated_data.get('flavor')
        executors.InstanceFlavorChangeExecutor().execute(instance, flavor=flavor, old_flavor_name=old_flavor_name)
        return response.Response({'status': _('change_flavor was scheduled')}, status=status.HTTP_202_ACCEPTED)

    def _can_change_flavor(instance):
        if (instance.state == models.Instance.States.OK and
                instance.runtime_state == models.Instance.RuntimeStates.ACTIVE):
            raise core_exceptions.IncorrectStateException(_('Please stop the instance before changing its flavor.'))

    change_flavor_serializer_class = serializers.InstanceFlavorChangeSerializer
    change_flavor_validators = [_can_change_flavor,
                                core_validators.StateValidator(models.Instance.States.OK),
                                core_validators.RuntimeStateValidator(models.Instance.RuntimeStates.SHUTOFF)]

    @decorators.detail_route(methods=['post'])
    def start(self, request, uuid=None):
        instance = self.get_object()
        executors.InstanceStartExecutor().execute(instance)
        return response.Response({'status': _('start was scheduled')}, status=status.HTTP_202_ACCEPTED)

    def _can_start_instance(instance):
        if (instance.state == models.Instance.States.OK and
                instance.runtime_state == models.Instance.RuntimeStates.ACTIVE):
            raise core_exceptions.IncorrectStateException(_('Instance is already active.'))

    start_validators = [_can_start_instance,
                        core_validators.StateValidator(models.Instance.States.OK),
                        core_validators.RuntimeStateValidator(models.Instance.RuntimeStates.SHUTOFF)]
    start_serializer_class = rf_serializers.Serializer

    @decorators.detail_route(methods=['post'])
    def stop(self, request, uuid=None):
        instance = self.get_object()
        executors.InstanceStopExecutor().execute(instance)
        return response.Response({'status': _('stop was scheduled')}, status=status.HTTP_202_ACCEPTED)

    def _can_stop_instance(instance):
        if (instance.state == models.Instance.States.OK and
                instance.runtime_state == models.Instance.RuntimeStates.SHUTOFF):
            raise core_exceptions.IncorrectStateException(_('Instance is already stopped.'))

    stop_validators = [_can_stop_instance,
                       core_validators.StateValidator(models.Instance.States.OK),
                       core_validators.RuntimeStateValidator(models.Instance.RuntimeStates.ACTIVE)]
    stop_serializer_class = rf_serializers.Serializer

    @decorators.detail_route(methods=['post'])
    def restart(self, request, uuid=None):
        instance = self.get_object()
        executors.InstanceRestartExecutor().execute(instance)
        return response.Response({'status': _('restart was scheduled')}, status=status.HTTP_202_ACCEPTED)

    def _can_restart_instance(instance):
        if (instance.state == models.Instance.States.OK and
                instance.runtime_state == models.Instance.RuntimeStates.SHUTOFF):
            raise core_exceptions.IncorrectStateException(_('Please start instance first.'))

    restart_validators = [_can_restart_instance,
                          core_validators.StateValidator(models.Instance.States.OK),
                          core_validators.RuntimeStateValidator(models.Instance.RuntimeStates.ACTIVE)]
    restart_serializer_class = rf_serializers.Serializer

    @decorators.detail_route(methods=['post'])
    def update_security_groups(self, request, uuid=None):
        instance = self.get_object()
        serializer = self.get_serializer(instance, data=request.data)
        serializer.is_valid(raise_exception=True)
        serializer.save()

        executors.InstanceUpdateSecurityGroupsExecutor().execute(instance)
        return response.Response({'status': _('security groups update was scheduled')}, status=status.HTTP_202_ACCEPTED)

    update_security_groups_validators = [core_validators.StateValidator(models.Instance.States.OK)]
    update_security_groups_serializer_class = serializers.InstanceSecurityGroupsUpdateSerializer

    @decorators.detail_route(methods=['post'])
    def backup(self, request, uuid=None):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        backup = serializer.save()

        executors.BackupCreateExecutor().execute(backup)
        return response.Response(serializer.data, status=status.HTTP_201_CREATED)

    backup_validators = [core_validators.StateValidator(models.Instance.States.OK)]
    backup_serializer_class = serializers.BackupSerializer

    @decorators.detail_route(methods=['post'])
    def create_backup_schedule(self, request, uuid=None):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        serializer.save()
        return response.Response(serializer.data, status=status.HTTP_201_CREATED)

    create_backup_schedule_validators = [core_validators.StateValidator(models.Instance.States.OK)]
    create_backup_schedule_serializer_class = serializers.BackupScheduleSerializer

    @decorators.detail_route(methods=['post'])
    def update_internal_ips_set(self, request, uuid=None):
        instance = self.get_object()
        serializer = self.get_serializer(instance, data=request.data)
        serializer.is_valid(raise_exception=True)
        serializer.save()

        executors.InstanceInternalIPsSetUpdateExecutor().execute(instance)
        return response.Response({'status': _('internal ips update was scheduled')}, status=status.HTTP_202_ACCEPTED)

    update_internal_ips_set_validators = [core_validators.StateValidator(models.Instance.States.OK)]
    update_internal_ips_set_serializer_class = serializers.InstanceInternalIPsSetUpdateSerializer

    @decorators.detail_route(methods=['get'])
    def internal_ips_set(self, request, uuid=None):
        instance = self.get_object()
        serializer = self.get_serializer(instance.internal_ips_set.all(), many=True)
        return response.Response(serializer.data, status=status.HTTP_200_OK)

    internal_ips_set_serializer_class = serializers.NestedInternalIPSerializer

    @decorators.detail_route(methods=['post'])
    def update_floating_ips(self, request, uuid=None):
        instance = self.get_object()
        serializer = self.get_serializer(instance, data=request.data)
        serializer.is_valid(raise_exception=True)
        serializer.save()

        executors.InstanceFloatingIPsUpdateExecutor().execute(instance)
        return response.Response({'status': _('floating ips update was scheduled')}, status=status.HTTP_202_ACCEPTED)

    update_floating_ips_validators = [core_validators.StateValidator(models.Instance.States.OK)]
    update_floating_ips_serializer_class = serializers.InstanceFloatingIPsUpdateSerializer

    @decorators.detail_route(methods=['get'])
    def floating_ips(self, request, uuid=None):
        instance = self.get_object()
        serializer = self.get_serializer(
            instance=instance.floating_ips.all(), queryset=models.FloatingIP.objects.all(), many=True)
        return response.Response(serializer.data, status=status.HTTP_200_OK)

    floating_ips_serializer_class = serializers.NestedFloatingIPSerializer

    importable_resources_backend_method = 'get_instances_for_import'
    importable_resources_serializer_class = serializers.InstanceImportableSerializer
    import_resource_serializer_class = serializers.InstanceImportSerializer
    import_resource_executor = executors.InstancePullExecutor

    @decorators.detail_route(methods=['get'])
    def console(self, request, uuid=None):
        instance = self.get_object()
        backend = instance.get_backend()
        try:
            url = backend.get_console_url(instance)
        except OpenStackBackendError as e:
            raise exceptions.ValidationError(e.message)

        return response.Response({'url': url}, status=status.HTTP_200_OK)

    console_validators = [core_validators.StateValidator(models.Instance.States.OK)]

    def check_permissions_for_console(request, view, instance=None):
        if not instance:
            return

        if request.user.is_staff:
            return

        if settings.WALDUR_OPENSTACK_TENANT['ALLOW_CUSTOMER_USERS_OPENSTACK_CONSOLE_ACCESS']:
            structure_permissions.is_administrator(request, view, instance)
        else:
            raise exceptions.PermissionDenied()

    console_permissions = [check_permissions_for_console]

    @decorators.detail_route(methods=['get'])
    def console_log(self, request, uuid=None):
        instance = self.get_object()
        backend = instance.get_backend()
        serializer = self.get_serializer(data=request.query_params)
        serializer.is_valid(raise_exception=True)
        length = serializer.validated_data.get('length')

        try:
            log = backend.get_console_output(instance, length)
        except OpenStackBackendError as e:
            raise exceptions.ValidationError(e.message)

        return response.Response(log, status=status.HTTP_200_OK)

    console_log_serializer_class = serializers.ConsoleLogSerializer
    console_log_permissions = [structure_permissions.is_administrator]


class BackupViewSet(structure_views.BaseResourceViewSet):
    queryset = models.Backup.objects.all().order_by('name')
    serializer_class = serializers.BackupSerializer
    filter_class = filters.BackupFilter
    disabled_actions = ['create']

    delete_executor = executors.BackupDeleteExecutor

    # method has to be overridden in order to avoid triggering of UpdateExecutor
    # which is a default action for all ResourceViewSet(s)
    def perform_update(self, serializer):
        serializer.save()

    @decorators.detail_route(methods=['post'])
    def restore(self, request, uuid=None):
        instance = self.get_object()
        serializer = self.get_serializer(instance, data=request.data)
        serializer.is_valid(raise_exception=True)
        backup_restoration = serializer.save()

        # Note that connected volumes will be linked with new marketplace.Resources by handler in openstack_marketplace
        structure_signals.resource_imported.send(
            sender=models.Instance,
            instance=backup_restoration.instance,
        )

        # It is assumed that SSH public key is already stored in OpenStack system volume.
        # Therefore we don't need to specify it explicitly for cloud init service.
        executors.InstanceCreateExecutor.execute(
            backup_restoration.instance,
            flavor=backup_restoration.flavor,
            is_heavy_task=True,
        )

        instance_serializer = serializers.InstanceSerializer(
            backup_restoration.instance, context={'request': self.request})
        return response.Response(instance_serializer.data, status=status.HTTP_201_CREATED)

    restore_validators = [core_validators.StateValidator(models.Backup.States.OK)]
    restore_serializer_class = serializers.BackupRestorationSerializer


class BaseScheduleViewSet(structure_views.BaseResourceViewSet):
    disabled_actions = ['create']

    # method has to be overridden in order to avoid triggering of UpdateExecutor
    # which is a default action for all ResourceViewSet(s)
    def perform_update(self, serializer):
        serializer.save()

    # method has to be overridden in order to avoid triggering of DeleteExecutor
    # which is a default action for all ResourceViewSet(s)
    def destroy(self, request, *args, **kwargs):
        resource = self.get_object()
        resource.delete()
        return response.Response(status=status.HTTP_204_NO_CONTENT)

    def list(self, request, *args, **kwargs):
        """
        For schedule to work, it should be activated - it's flag is_active set to true. If it's not, it won't be used
        for triggering the next operations. Schedule will be deactivated if operation fails.

        - **retention time** is a duration in days during which resource is preserved.
        - **maximal_number_of_resources** is a maximal number of active resources connected to this schedule.
        - **schedule** is a resource schedule defined in a cron format.
        - **timezone** is used for calculating next run of the resource schedule (optional).

        A schedule can be it two states: active or not. Non-active states are not used for scheduling the new tasks.
        Only users with write access to schedule resource can activate or deactivate a schedule.
        """
        return super(BaseScheduleViewSet, self).list(self, request, *args, **kwargs)

    def _is_schedule_active(resource_schedule):
        if resource_schedule.is_active:
            raise core_exceptions.IncorrectStateException(_('Resource schedule is already activated.'))

    @decorators.detail_route(methods=['post'])
    def activate(self, request, uuid):
        """
        Activate a resource schedule. Note that
        if a schedule is already active, this will result in **409 CONFLICT** code.
        """
        schedule = self.get_object()
        schedule.is_active = True
        schedule.error_message = ''
        schedule.save()
        return response.Response({'status': _('A schedule was activated')})

    activate_validators = [_is_schedule_active]

    def _is_schedule_deactived(resource_schedule):
        if not resource_schedule.is_active:
            raise core_exceptions.IncorrectStateException(_('A schedule is already deactivated.'))

    @decorators.detail_route(methods=['post'])
    def deactivate(self, request, uuid):
        """
        Deactivate a resource schedule. Note that
        if a schedule was already deactivated, this will result in **409 CONFLICT** code.
        """
        schedule = self.get_object()
        schedule.is_active = False
        schedule.save()
        return response.Response({'status': _('Backup schedule was deactivated')})

    deactivate_validators = [_is_schedule_deactived]


class BackupScheduleViewSet(BaseScheduleViewSet):
    queryset = models.BackupSchedule.objects.all().order_by('name')
    serializer_class = serializers.BackupScheduleSerializer
    filter_class = filters.BackupScheduleFilter


class SnapshotScheduleViewSet(BaseScheduleViewSet):
    queryset = models.SnapshotSchedule.objects.all().order_by('name')
    serializer_class = serializers.SnapshotScheduleSerializer
    filter_class = filters.SnapshotScheduleFilter


class SharedSettingsBaseView(generics.GenericAPIView):
    def get_private_settings(self):
        service_settings_uuid = self.request.query_params.get('service_settings_uuid')
        if not service_settings_uuid or not core_utils.is_uuid_like(service_settings_uuid):
            return structure_models.ServiceSettings.objects.none()

        queryset = structure_models.ServiceSettings.objects.filter(type=OpenStackConfig.service_name)
        try:
            shared_settings = queryset.get(uuid=service_settings_uuid)
        except structure_models.ServiceSettings.DoesNotExist:
            return structure_models.ServiceSettings.objects.none()

        tenants = openstack_models.Tenant.objects.filter(service_project_link__service__settings=shared_settings)
        if tenants:
            return structure_models.ServiceSettings.objects.filter(scope__in=tenants)
        else:
            return structure_models.ServiceSettings.objects.none()

    def get(self, request, *args, **kwargs):
        page = self.paginate_queryset(self.get_queryset())
        serializer = self.get_serializer(page, many=True)
        return self.get_paginated_response(serializer.data)


class SharedSettingsInstances(SharedSettingsBaseView):

    serializer_class = serializers.InstanceSerializer

    def get_queryset(self):
        private_settings = self.get_private_settings()
        return models.Instance.objects\
            .order_by('service_project_link__project__customer__name')\
            .filter(service_project_link__service__settings__in=private_settings)


class SharedSettingsCustomers(SharedSettingsBaseView):

    serializer_class = serializers.SharedSettingsCustomerSerializer

    def get_queryset(self):
        private_settings = self.get_private_settings()
        vms = models.Instance.objects.filter(
            service_project_link__service__settings__in=private_settings,
            project__customer=OuterRef('pk')
        ).annotate(count=Count('*')).values('count')

        # Workaround for Django bug:
        # https://code.djangoproject.com/ticket/28296
        # It allows to remove extra GROUP BY clause from the subquery.
        vms.query.group_by = []

        vm_count = Subquery(vms[:1], output_field=IntegerField())
        return structure_models.Customer.objects.filter(
            pk__in=private_settings.values('customer')
        ).annotate(vm_count=vm_count)


class VolumeTypeViewSet(structure_views.BaseServicePropertyViewSet):
    queryset = models.VolumeType.objects.all().order_by('settings', 'name')
    serializer_class = serializers.VolumeTypeSerializer
    lookup_field = 'uuid'
    filter_class = filters.VolumeTypeFilter


class VolumeAvailabilityZoneViewSet(structure_views.BaseServicePropertyViewSet):
    queryset = models.VolumeAvailabilityZone.objects.all().order_by('settings', 'name')
    serializer_class = serializers.VolumeAvailabilityZoneSerializer
    lookup_field = 'uuid'
    filter_class = filters.VolumeAvailabilityZoneFilter



import copy

from waldur_mastermind.marketplace import serializers as marketplace_serializers


class BookingResourceSerializer(marketplace_serializers.ResourceSerializer):
    class Meta(marketplace_serializers.ResourceSerializer.Meta):
        view_name = 'booking-resource-detail'
        fields = marketplace_serializers.ResourceSerializer.Meta.fields + ('url',)
        extra_kwargs = copy.copy(marketplace_serializers.ResourceSerializer.Meta.extra_kwargs)
        extra_kwargs['url'] = {'lookup_field': 'uuid', 'read_only': True}



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.conf import settings
from django.core import validators
from django.db import migrations, models


def update_users_tokens_lifetime(apps, schema_editor):
    if settings.WALDUR_CORE['TOKEN_LIFETIME']:
        seconds = settings.WALDUR_CORE['TOKEN_LIFETIME'].total_seconds()
        User = apps.get_model('core', 'User')
        User.objects.update(
            token_lifetime=int(seconds)
        )


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0006_user_is_support'),
    ]

    operations = [
        migrations.AddField(
            model_name='user',
            name='token_lifetime',
            field=models.PositiveIntegerField(help_text='Token lifetime in seconds.', null=True, validators=[validators.MinValueValidator(60)]),
        ),
        migrations.RunPython(update_users_tokens_lifetime),
    ]



from waldur_azure import views as azure_views
from waldur_mastermind.marketplace import processors


class VirtualMachineCreateProcessor(processors.BaseCreateResourceProcessor):
    viewset = azure_views.VirtualMachineViewSet
    fields = (
        'name',
        'description',
        'size',
        'image',
        'location',
    )


class VirtualMachineDeleteProcessor(processors.DeleteResourceProcessor):
    viewset = azure_views.VirtualMachineViewSet


class SQLServerCreateProcessor(processors.BaseCreateResourceProcessor):
    viewset = azure_views.SQLServerViewSet

    fields = (
        'name',
        'description',
        'location',
    )


class SQLServerDeleteProcessor(processors.DeleteResourceProcessor):
    viewset = azure_views.SQLServerViewSet



from waldur_core.logging.loggers import EventLogger, event_logger
from waldur_core.structure.models import Project, Customer


class PriceEstimateEventLogger(EventLogger):

    class Meta:
        event_types = ('project_price_limit_updated', 'customer_price_limit_updated')
        event_groups = {
            'projects': ['project_price_limit_updated'],
            'customers': ['customer_price_limit_updated'],
        }

    @staticmethod
    def get_scopes(event_context):
        scope = event_context['scope']
        if isinstance(scope, Project):
            return {scope, scope.customer}
        elif isinstance(scope, Customer):
            return {scope}
        else:
            return set()


event_logger.register('price_estimate', PriceEstimateEventLogger)



from . import views


def register_in(router):
    router.register(r'slurm', views.SlurmServiceViewSet, base_name='slurm')
    router.register(r'slurm-service-project-link', views.SlurmServiceProjectLinkViewSet,
                    base_name='slurm-spl')
    router.register(r'slurm-allocation', views.AllocationViewSet, base_name='slurm-allocation')
    router.register(r'slurm-allocation-usage', views.AllocationUsageViewSet, base_name='slurm-allocation-usage')



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


def pull_instance_tenant(apps, schema_editor):
    Instance = apps.get_model('openstack', 'Instance')
    for instance in Instance.objects.all():
        instance.tenant = instance.service_project_link.tenants.first()
        instance.save()


class Migration(migrations.Migration):

    dependencies = [
        ('openstack', '0006_backups_restorations'),
    ]

    operations = [
        migrations.AddField(
            model_name='instance',
            name='tenant',
            field=models.ForeignKey(related_name='instances', to='openstack.Tenant', null=True),
        ),
        migrations.RunPython(pull_instance_tenant, elidable=True),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.20 on 2019-06-19 17:56
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0086_cartitem_project'),
    ]

    operations = [
        migrations.AddField(
            model_name='componentusage',
            name='billing_period',
            field=models.DateField(null=True),
        ),
        migrations.AlterField(
            model_name='componentusage',
            name='date',
            field=models.DateTimeField(),
        ),
        migrations.AlterUniqueTogether(
            name='componentusage',
            unique_together=set([('resource', 'component', 'plan_period', 'billing_period')]),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import uuid

from django.db import models, migrations


def init_is_closed_attribute(apps, schema_editor):
    Alert = apps.get_model('logging', 'Alert')
    for alert in Alert.objects.all():
        if alert.closed is None:
            # Delete duplicates to prevent integrity errors
            (Alert.objects
                .filter(object_id=alert.object_id,
                        content_type=alert.content_type,
                        alert_type=alert.alert_type,
                        closed__isnull=True)
                .delete())
        else:
            alert.is_closed = uuid.uuid4().hex
            alert.save()


class Migration(migrations.Migration):

    #replaces = [('logging', '0003_add_alert_unique_together_constraint')]

    dependencies = [
        ('logging', '0004_index_alert_type'),
    ]

    operations = [
        migrations.AddField(
            model_name='alert',
            name='is_closed',
            field=models.CharField(max_length=32, blank=True),
            preserve_default=True,
        ),
        migrations.RunPython(init_is_closed_attribute),
        migrations.AlterUniqueTogether(
            name='alert',
            unique_together=set([('content_type', 'object_id', 'alert_type', 'is_closed')]),
        ),
    ]



from django.apps import AppConfig
from django.db.models import signals


class InvoiceConfig(AppConfig):
    name = 'waldur_mastermind.invoices'
    verbose_name = 'Invoices'

    def ready(self):
        from waldur_core.core import signals as core_signals
        from waldur_core.structure import models as structure_models

        from . import handlers, models

        signals.pre_save.connect(
            handlers.set_tax_percent_on_invoice_creation,
            sender=models.Invoice,
            dispatch_uid='waldur_mastermind.invoices.set_tax_percent_on_invoice_creation',
        )

        signals.post_save.connect(
            handlers.log_invoice_state_transition,
            sender=models.Invoice,
            dispatch_uid='waldur_mastermind.invoices.log_invoice_state_transition',
        )

        signals.post_save.connect(
            handlers.emit_invoice_created_event,
            sender=models.Invoice,
            dispatch_uid='waldur_mastermind.invoices.emit_invoice_created_event',
        )

        signals.post_save.connect(
            handlers.update_invoice_pdf,
            sender=models.Invoice,
            dispatch_uid='waldur_mastermind.invoices.update_invoice_pdf',
        )

        signals.post_save.connect(
            handlers.set_project_name_on_invoice_item_creation,
            sender=models.InvoiceItem,
            dispatch_uid='waldur_mastermind.invoices.set_project_name_on_invoice_item_creation',
        )

        signals.post_save.connect(
            handlers.update_current_cost_when_invoice_item_is_updated,
            sender=models.InvoiceItem,
            dispatch_uid='waldur_mastermind.invoices.update_current_cost_when_invoice_item_is_updated_%s',
        )

        signals.post_delete.connect(
            handlers.update_current_cost_when_invoice_item_is_deleted,
            sender=models.InvoiceItem,
            dispatch_uid='waldur_mastermind.invoices.update_current_cost_when_invoice_item_is_deleted',
        )

        signals.post_save.connect(
            handlers.update_invoice_item_on_project_name_update,
            sender=structure_models.Project,
            dispatch_uid='waldur_mastermind.invoices.update_invoice_item_on_project_name_update',
        )

        core_signals.pre_delete_validate.connect(
            handlers.prevent_deletion_of_customer_with_invoice,
            sender=structure_models.Customer,
            dispatch_uid='waldur_mastermind.invoices.prevent_deletion_of_customer_with_invoice',
        )

        signals.post_save.connect(
            handlers.adjust_invoice_items_for_downtime,
            sender=models.ServiceDowntime,
            dispatch_uid='waldur_mastermind.invoices.adjust_invoice_items_for_downtime',
        )



from django.apps import AppConfig
from django.db.models import signals


class JiraConfig(AppConfig):
    name = 'waldur_jira'
    verbose_name = 'JIRA'
    service_name = 'JIRA'

    def ready(self):
        from waldur_core.quotas import fields as quota_fields
        from waldur_core.structure import SupportedServices
        from waldur_core.structure import models as structure_models
        from waldur_core.structure.signals import resource_imported

        from . import handlers
        from .backend import JiraBackend
        SupportedServices.register_backend(JiraBackend)

        Issue = self.get_model('Issue')
        Comment = self.get_model('Comment')
        Project = self.get_model('Project')

        structure_models.Project.add_quota_field(
            name='nc_jira_project_count',
            quota_field=quota_fields.CounterQuotaField(
                target_models=lambda: [Project],
                path_to_scope='service_project_link.project',
            )
        )

        resource_imported.connect(
            handlers.import_project_issues,
            sender=Project,
            dispatch_uid='waldur_jira.handlers.import_project_issues',
        )

        signals.post_save.connect(
            handlers.log_issue_save,
            sender=Issue,
            dispatch_uid='waldur_jira.handlers.log_issue_save',
        )

        signals.post_delete.connect(
            handlers.log_issue_delete,
            sender=Issue,
            dispatch_uid='waldur_jira.handlers.log_issue_delete',
        )

        signals.post_save.connect(
            handlers.log_comment_save,
            sender=Comment,
            dispatch_uid='waldur_jira.handlers.log_comment_save',
        )

        signals.post_delete.connect(
            handlers.log_comment_delete,
            sender=Comment,
            dispatch_uid='waldur_jira.handlers.log_comment_delete',
        )



from __future__ import unicode_literals

import datetime
import logging

import jwt
from django.core.exceptions import ValidationError, ObjectDoesNotExist
from django.db import transaction
from django.db.models import OuterRef, Subquery, Count, IntegerField
from django.utils import timezone
from django.utils.translation import ugettext_lazy as _
from rest_framework import exceptions as rf_exceptions
from rest_framework import serializers
from rest_framework.reverse import reverse

from waldur_core.core import serializers as core_serializers
from waldur_core.core import signals as core_signals
from waldur_core.core import utils as core_utils
from waldur_core.core.fields import NaturalChoiceField
from waldur_core.core.serializers import GenericRelatedField
from waldur_core.media.serializers import ProtectedMediaSerializerMixin, ProtectedFileField
from waldur_core.quotas.serializers import BasicQuotaSerializer
from waldur_core.structure import models as structure_models, SupportedServices
from waldur_core.structure import permissions as structure_permissions
from waldur_core.structure import serializers as structure_serializers
from waldur_core.structure.managers import filter_queryset_for_user
from waldur_core.structure.tasks import connect_shared_settings
from waldur_mastermind.common.serializers import validate_options
from waldur_mastermind.common import exceptions
from waldur_mastermind.marketplace.permissions import check_availability_of_auto_approving
from waldur_mastermind.marketplace.plugins import manager
from waldur_mastermind.marketplace.utils import validate_order_item, validate_limits, create_offering_components
from waldur_mastermind.support import serializers as support_serializers
from waldur_mastermind.marketplace.processors import CreateResourceProcessor

from . import models, attribute_types, plugins, permissions, tasks

logger = logging.getLogger(__name__)


class ServiceProviderSerializer(core_serializers.AugmentedSerializerMixin,
                                serializers.HyperlinkedModelSerializer):
    class Meta(object):
        model = models.ServiceProvider
        fields = (
            'url', 'uuid', 'created', 'description', 'enable_notifications',
            'customer', 'customer_name', 'customer_uuid', 'customer_image',
            'customer_abbreviation', 'customer_native_name',
        )
        related_paths = {
            'customer': ('uuid', 'name', 'native_name', 'abbreviation')
        }
        protected_fields = ('customer',)
        extra_kwargs = {
            'url': {'lookup_field': 'uuid', 'view_name': 'marketplace-service-provider-detail'},
            'customer': {'lookup_field': 'uuid'},
        }

    customer_image = ProtectedFileField(source='customer.image', read_only=True)

    def validate(self, attrs):
        if not self.instance:
            permissions.can_register_service_provider(self.context['request'], attrs['customer'])
        return attrs


class NestedAttributeOptionSerializer(serializers.ModelSerializer):
    class Meta(object):
        model = models.AttributeOption
        fields = ('key', 'title')


class NestedAttributeSerializer(serializers.ModelSerializer):
    options = NestedAttributeOptionSerializer(many=True)

    class Meta(object):
        model = models.Attribute
        fields = ('key', 'title', 'type', 'options', 'required', 'default')


class NestedSectionSerializer(serializers.ModelSerializer):
    attributes = NestedAttributeSerializer(many=True, read_only=True)

    class Meta(object):
        model = models.Section
        fields = ('key', 'title', 'attributes', 'is_standalone')


class NestedColumnSerializer(serializers.ModelSerializer):
    class Meta(object):
        model = models.CategoryColumn
        fields = ('index', 'title', 'attribute', 'widget')


class CategoryComponentSerializer(serializers.ModelSerializer):
    class Meta(object):
        model = models.CategoryComponent
        fields = ('type', 'name', 'description', 'measured_unit')


class CategorySerializer(ProtectedMediaSerializerMixin,
                         core_serializers.AugmentedSerializerMixin,
                         core_serializers.RestrictedSerializerMixin,
                         serializers.HyperlinkedModelSerializer):
    offering_count = serializers.ReadOnlyField()
    sections = NestedSectionSerializer(many=True, read_only=True)
    columns = NestedColumnSerializer(many=True, read_only=True)
    components = CategoryComponentSerializer(many=True, read_only=True)

    @staticmethod
    def eager_load(queryset, request):
        offerings = models.Offering.objects \
            .filter(state=models.Offering.States.ACTIVE) \
            .filter(category=OuterRef('pk')) \
            .filter_for_user(request.user)

        allowed_customer_uuid = request.query_params.get('allowed_customer_uuid')
        if allowed_customer_uuid and core_utils.is_uuid_like(allowed_customer_uuid):
            offerings = offerings.filter_for_customer(allowed_customer_uuid)

        project_uuid = request.query_params.get('project_uuid')
        if project_uuid and core_utils.is_uuid_like(project_uuid):
            offerings = offerings.filter_for_project(project_uuid)

        offerings = offerings \
            .annotate(count=Count('*'))\
            .values('count')

        # Workaround for Django bug:
        # https://code.djangoproject.com/ticket/28296
        # It allows to remove extra GROUP BY clause from the subquery.
        offerings.query.group_by = []

        offering_count = Subquery(offerings[:1], output_field=IntegerField())
        queryset = queryset.annotate(offering_count=offering_count)
        return queryset.prefetch_related('sections', 'sections__attributes')

    class Meta(object):
        model = models.Category
        fields = ('url', 'uuid', 'title', 'description', 'icon', 'offering_count',
                  'sections', 'columns', 'components')
        extra_kwargs = {
            'url': {'lookup_field': 'uuid', 'view_name': 'marketplace-category-detail'},
        }


PriceSerializer = serializers.DecimalField(
    min_value=0,
    max_digits=models.PlanComponent.PRICE_MAX_DIGITS,
    decimal_places=models.PlanComponent.PRICE_DECIMAL_PLACES,
)


class BasePlanSerializer(core_serializers.AugmentedSerializerMixin,
                         serializers.HyperlinkedModelSerializer):
    prices = serializers.DictField(child=PriceSerializer, write_only=True, required=False)
    quotas = serializers.DictField(child=serializers.IntegerField(min_value=0),
                                   write_only=True, required=False)

    class Meta(object):
        model = models.Plan
        fields = ('url', 'uuid', 'name', 'description',
                  'article_code', 'product_code',
                  'prices', 'quotas', 'max_amount', 'archived', 'is_active',
                  'unit_price', 'unit', 'init_price', 'switch_price')
        read_ony_fields = ('unit_price', 'archived')
        extra_kwargs = {
            'url': {'lookup_field': 'uuid', 'view_name': 'marketplace-plan-detail'},
        }

    def get_fields(self):
        fields = super(BasePlanSerializer, self).get_fields()
        method = self.context['view'].request.method
        if method == 'GET':
            fields['prices'] = serializers.SerializerMethodField()
            fields['quotas'] = serializers.SerializerMethodField()
        return fields

    def get_prices(self, plan):
        return {item.component.type: item.price for item in plan.components.all()}

    def get_quotas(self, plan):
        return {item.component.type: item.amount for item in plan.components.all()}


class PlanDetailsSerializer(BasePlanSerializer):
    class Meta(BasePlanSerializer.Meta):
        model = models.Plan
        fields = BasePlanSerializer.Meta.fields + ('offering',)
        protected_fields = ('offering',)
        extra_kwargs = {
            'url': {'lookup_field': 'uuid', 'view_name': 'marketplace-plan-detail'},
            'offering': {'lookup_field': 'uuid', 'view_name': 'marketplace-offering-detail'},
        }

    def validate(self, attrs):
        if not self.instance:
            structure_permissions.is_owner(self.context['request'], None, attrs['offering'].customer)

        return attrs


class PlanUsageRequestSerializer(serializers.Serializer):
    offering_uuid = serializers.UUIDField(required=False)
    customer_provider_uuid = serializers.UUIDField(required=False)
    o = serializers.ChoiceField(choices=(
        'usage', 'limit', 'remaining',
        '-usage', '-limit', '-remaining',
    ), required=False)


class PlanUsageResponseSerializer(serializers.Serializer):
    plan_uuid = serializers.ReadOnlyField(source='uuid')
    plan_name = serializers.ReadOnlyField(source='name')

    limit = serializers.ReadOnlyField()
    usage = serializers.ReadOnlyField()
    remaining = serializers.ReadOnlyField()

    offering_uuid = serializers.ReadOnlyField(source='offering.uuid')
    offering_name = serializers.ReadOnlyField(source='offering.name')

    customer_provider_uuid = serializers.ReadOnlyField(source='offering.customer.uuid')
    customer_provider_name = serializers.ReadOnlyField(source='offering.customer.name')


class NestedScreenshotSerializer(ProtectedMediaSerializerMixin, serializers.ModelSerializer):
    class Meta(object):
        model = models.Screenshot
        fields = ('name', 'description', 'image', 'thumbnail')


class NestedOfferingFileSerializer(ProtectedMediaSerializerMixin, serializers.ModelSerializer):
    class Meta(object):
        model = models.OfferingFile
        fields = ('name', 'created', 'file',)


class ScreenshotSerializer(ProtectedMediaSerializerMixin,
                           core_serializers.AugmentedSerializerMixin,
                           serializers.HyperlinkedModelSerializer):

    class Meta(object):
        model = models.Screenshot
        fields = ('url', 'uuid', 'name', 'description', 'image', 'thumbnail', 'offering')
        protected_fields = ('offering', 'image')
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
            'offering': {'lookup_field': 'uuid', 'view_name': 'marketplace-offering-detail'},
        }

    def validate(self, attrs):
        if not self.instance:
            structure_permissions.is_owner(self.context['request'], None, attrs['offering'].customer)
        return attrs


FIELD_TYPES = (
    'boolean',
    'integer',
    'money',
    'string',
    'text',
    'html_text',
    'select_string',
    'select_openstack_tenant',
    'select_openstack_instance',
    'select_multiple_openstack_instances',
    'date',
    'time',
)


class DefaultField(serializers.Field):
    def to_internal_value(self, data):
        return data


class OptionFieldSerializer(serializers.Serializer):
    type = serializers.ChoiceField(choices=FIELD_TYPES)
    label = serializers.CharField()
    help_text = serializers.CharField(required=False)
    required = serializers.BooleanField(default=False)
    choices = serializers.ListField(child=serializers.CharField(), required=False)
    default = DefaultField(required=False)
    min = serializers.IntegerField(required=False)
    max = serializers.IntegerField(required=False)


class OfferingOptionsSerializer(serializers.Serializer):
    order = serializers.ListField(child=serializers.CharField())
    options = serializers.DictField(child=OptionFieldSerializer())


class OfferingComponentSerializer(serializers.ModelSerializer):
    class Meta(object):
        model = models.OfferingComponent
        fields = ('billing_type', 'type', 'name', 'description', 'measured_unit',
                  'limit_period', 'limit_amount', 'disable_quotas', 'product_code', 'article_code',
                  'max_value', 'min_value')
        extra_kwargs = {
            'billing_type': {'required': True},
        }


class OfferingDetailsSerializer(ProtectedMediaSerializerMixin,
                                core_serializers.AugmentedSerializerMixin,
                                core_serializers.RestrictedSerializerMixin,
                                serializers.HyperlinkedModelSerializer):

    attributes = serializers.JSONField(required=False)
    options = serializers.JSONField(required=False)
    components = OfferingComponentSerializer(required=False, many=True)
    geolocations = core_serializers.GeoLocationField(required=False)
    order_item_count = serializers.SerializerMethodField()
    plans = BasePlanSerializer(many=True, required=False)
    screenshots = NestedScreenshotSerializer(many=True, read_only=True)
    state = serializers.ReadOnlyField(source='get_state_display')
    scope = GenericRelatedField(read_only=True)
    scope_uuid = serializers.ReadOnlyField(source='scope.uuid')
    files = NestedOfferingFileSerializer(many=True, read_only=True)
    quotas = serializers.SerializerMethodField()

    class Meta(object):
        model = models.Offering
        fields = ('url', 'uuid', 'created', 'name', 'description', 'full_description', 'terms_of_service',
                  'customer', 'customer_uuid', 'customer_name',
                  'category', 'category_uuid', 'category_title',
                  'rating', 'attributes', 'options', 'components', 'geolocations',
                  'state', 'native_name', 'native_description', 'vendor_details',
                  'thumbnail', 'order_item_count', 'plans', 'screenshots', 'type', 'shared', 'billable',
                  'scope', 'scope_uuid', 'files', 'quotas', 'paused_reason')
        related_paths = {
            'customer': ('uuid', 'name'),
            'category': ('uuid', 'title'),
        }
        protected_fields = ('customer', 'type')
        read_only_fields = ('state', 'paused_reason')
        extra_kwargs = {
            'url': {'lookup_field': 'uuid', 'view_name': 'marketplace-offering-detail'},
            'customer': {'lookup_field': 'uuid', 'view_name': 'customer-detail'},
            'category': {'lookup_field': 'uuid', 'view_name': 'marketplace-category-detail'},
        }

    def get_order_item_count(self, offering):
        try:
            return offering.quotas.get(name='order_item_count').usage
        except ObjectDoesNotExist:
            return 0

    def get_quotas(self, offering):
        if offering.scope and hasattr(offering.scope, 'quotas'):
            return BasicQuotaSerializer(offering.scope.quotas, many=True, context=self.context).data


class OfferingComponentLimitSerializer(serializers.Serializer):
    min = serializers.IntegerField(min_value=0)
    max = serializers.IntegerField(min_value=0)


class OfferingModifySerializer(OfferingDetailsSerializer):

    class Meta(OfferingDetailsSerializer.Meta):
        model = models.Offering
        fields = OfferingDetailsSerializer.Meta.fields + ('limits',)

    limits = serializers.DictField(child=OfferingComponentLimitSerializer(),
                                   write_only=True, required=False)

    def validate(self, attrs):
        if not self.instance:
            structure_permissions.is_owner(self.context['request'], None, attrs['customer'])

        self._validate_attributes(attrs)
        self._validate_plans(attrs)

        return attrs

    def validate_type(self, offering_type):
        if offering_type not in plugins.manager.backends.keys():
            raise rf_exceptions.ValidationError(_('Invalid value.'))
        return offering_type

    def validate_terms_of_service(self, value):
        if value:
            value = value.strip()
        return value

    def _validate_attributes(self, attrs):
        category = attrs.get('category')
        if category is None and self.instance:
            category = self.instance.category

        attributes = attrs.get('attributes')
        if attributes is not None and not isinstance(attributes, dict):
            raise rf_exceptions.ValidationError({
                'attributes': _('Dictionary is expected.'),
            })

        if attributes is None and self.instance:
            attributes = self.instance.attributes

        if attributes is None:
            attributes = dict()

        category_attributes = models.Attribute.objects.filter(section__category=category)
        required_attributes = category_attributes.filter(required=True).values_list('key', flat=True)
        missing_attributes = set(required_attributes) - set(attributes.keys())

        if missing_attributes:
            raise rf_exceptions.ValidationError({
                'attributes': _('These attributes are required: %s' % ', '.join(sorted(missing_attributes)))
            })

        for attribute in category_attributes:
            value = attributes.get(attribute.key)
            if value is None:
                # Use default attribute value if it is defined
                if attribute.default is not None:
                    attributes[attribute.key] = attribute.default
                continue

            validator = attribute_types.get_attribute_type(attribute.type)
            if not validator:
                continue

            try:
                validator.validate(value, list(attribute.options.values_list('key', flat=True)))
            except ValidationError as e:
                raise rf_exceptions.ValidationError({attribute.key: e.message})

    def validate_options(self, options):
        serializer = OfferingOptionsSerializer(data=options)
        serializer.is_valid(raise_exception=True)
        return serializer.validated_data

    def _validate_plans(self, attrs):
        custom_components = attrs.get('components')
        if not custom_components and self.instance:
            custom_components = self.instance.components.all().values()

        offering_type = attrs.get('type', getattr(self.instance, 'type', None))
        builtin_components = plugins.manager.get_components(offering_type)

        valid_types = set()
        fixed_types = set()

        if builtin_components and attrs.get('components'):
            raise serializers.ValidationError({
                'components': _('Extra components are not allowed.')
            })

        elif builtin_components:
            valid_types = {component.type for component in builtin_components}
            fixed_types = {component.type
                           for component in plugins.manager.get_components(offering_type)
                           if component.billing_type == models.OfferingComponent.BillingTypes.FIXED}

        elif custom_components:
            valid_types = {component['type'] for component in custom_components}
            fixed_types = {component['type'] for component in custom_components
                           if component['billing_type'] == models.OfferingComponent.BillingTypes.FIXED}

        for plan in attrs.get('plans', []):
            plan_name = plan.get('name')

            prices = plan.get('prices', {})
            invalid_components = ', '.join(sorted(set(prices.keys()) - valid_types))
            if invalid_components:
                raise serializers.ValidationError({
                    'plans': _('Invalid price components %s in plan "%s".') % (invalid_components, plan_name)
                })

            quotas = plan.get('quotas', {})
            invalid_components = ', '.join(sorted(set(quotas.keys()) - fixed_types))
            if invalid_components:
                raise serializers.ValidationError({
                    'plans': _('Invalid quota components %s in plan "%s".') % (invalid_components, plan_name),
                })

            plan['unit_price'] = sum(
                prices.get(component, 0) * quotas.get(component, 0)
                for component in fixed_types
            )

    def _create_plan(self, offering, plan_data, components):
        quotas = plan_data.pop('quotas', {})
        prices = plan_data.pop('prices', {})
        plan = models.Plan.objects.create(offering=offering, **plan_data)

        for name, component in components.items():
            models.PlanComponent.objects.create(
                plan=plan,
                component=component,
                amount=quotas.get(name) or 0,
                price=prices.get(name) or 0,
            )

    def _create_plans(self, offering, plans):
        components = {component.type: component for component in offering.components.all()}
        for plan_data in plans:
            self._create_plan(offering, plan_data, components)

    def _update_limits(self, offering, limits):
        for key, values in limits.items():
            models.OfferingComponent.objects.filter(
                offering=offering,
                type=key
            ).update(
                min_value=values['min'],
                max_value=values['max']
            )


class OfferingCreateSerializer(OfferingModifySerializer):
    class Meta(OfferingModifySerializer.Meta):
        fields = OfferingModifySerializer.Meta.fields + ('service_attributes',)

    service_attributes = serializers.JSONField(required=False, write_only=True)

    def validate_plans(self, plans):
        if len(plans) < 1:
            raise serializers.ValidationError({
                'plans': _('At least one plan should be specified.')
            })
        return plans

    @transaction.atomic
    def create(self, validated_data):
        limits = validated_data.pop('limits', {})
        plans = validated_data.pop('plans', [])
        custom_components = validated_data.pop('components', [])

        offering_type = validated_data.get('type')
        service_type = plugins.manager.get_service_type(offering_type)
        if service_type:
            validated_data = self._create_service(service_type, validated_data)

        offering = super(OfferingCreateSerializer, self).create(validated_data)
        create_offering_components(offering, custom_components)
        if limits:
            self._update_limits(offering, limits)
        self._create_plans(offering, plans)

        return offering

    def _create_service(self, service_type, validated_data):
        """
        Marketplace offering model does not accept service_attributes field as is,
        therefore we should remove it from validated_data and create service settings object.
        Then we need to specify created object and offering's scope.
        """
        name = validated_data['name']
        service_attributes = validated_data.pop('service_attributes', {})
        if not service_attributes:
            raise ValidationError({
                'service_attributes': _('This field is required.')
            })
        payload = dict(
            name=name,
            # It is expected that customer URL is passed to the service settings serializer
            customer=self.initial_data['customer'],
            type=service_type,
            **service_attributes
        )
        serializer_class = SupportedServices.get_service_serializer_for_key(service_type)
        serializer = serializer_class(data=payload, context=self.context)
        serializer.is_valid(raise_exception=True)
        service = serializer.save()
        # Usually we don't allow users to create new shared service settings via REST API.
        # That's shared flag is marked as read-only in service settings serializer.
        # But shared offering should be created with shared service settings.
        # That's why we set it to shared only after service settings object is created.
        if validated_data.get('shared'):
            service.settings.shared = True
            service.settings.save()
            # Usually connect shared settings task is called when service is created.
            # But as we set shared flag after serializer has been executed,
            # we need to connect shared settings manually.
            connect_shared_settings(service.settings)
        validated_data['scope'] = service.settings
        return validated_data


class OfferingPauseSerializer(serializers.ModelSerializer):
    class Meta(object):
        model = models.Offering
        fields = ['paused_reason']


class PlanUpdateSerializer(BasePlanSerializer):

    class Meta(BasePlanSerializer.Meta):
        extra_kwargs = {
            'uuid': {'read_only': False},
        }


class OfferingUpdateSerializer(OfferingModifySerializer):

    plans = PlanUpdateSerializer(many=True, required=False, write_only=True)

    def _update_components(self, instance, components):
        resources_exist = models.Resource.objects.filter(offering=instance).exists()

        old_components = {
            component.type: component
            for component in instance.components.all()
        }

        new_components = {
            component['type']: models.OfferingComponent(offering=instance, **component)
            for component in components
        }

        removed_components = set(old_components.keys()) - set(new_components.keys())
        added_components = set(new_components.keys()) - set(old_components.keys())
        updated_components = set(new_components.keys()) & set(old_components.keys())

        builtin_components = plugins.manager.get_components(self.instance.type)
        valid_types = {component.type for component in builtin_components}

        if removed_components & valid_types:
            raise serializers.ValidationError({
                'components': _('These components cannot be removed because they are builtin: %s') %
                ', '.join(removed_components & valid_types)
            })

        if updated_components & valid_types:
            raise serializers.ValidationError({
                'components': _('These components cannot be updated because they are builtin: %s') %
                ', '.join(updated_components & valid_types)
            })

        if removed_components:
            if resources_exist:
                raise serializers.ValidationError({
                    'components': _('These components cannot be removed because they are already used: %s') %
                    ', '.join(removed_components)
                })
            else:
                models.OfferingComponent.objects.filter(type__in=removed_components).delete()

        for key in added_components:
            new_components[key].save()

        COMPONENT_KEYS = (
            'name', 'description',
            'billing_type', 'measured_unit',
            'limit_period', 'limit_amount', 'disable_quotas',
            'product_code', 'article_code',
        )

        for component_key in updated_components:
            new_component = new_components[component_key]
            old_component = old_components[component_key]
            for key in COMPONENT_KEYS:
                setattr(old_component, key, getattr(new_component, key))
            old_component.save()

    def _update_plan_components(self, old_plan, new_plan):
        new_quotas = new_plan.get('quotas', {})
        new_prices = new_plan.get('prices', {})

        new_keys = set(new_quotas.keys()) | set(new_prices.keys())
        old_keys = set(old_plan.components.values_list('component__type', flat=True))

        for key in new_keys - old_keys:
            component = old_plan.offering.components.get(type=key)
            models.PlanComponent.objects.create(plan=old_plan, component=component)

    def _update_quotas(self, old_plan, new_plan):
        new_quotas = new_plan.get('quotas', {})
        new_prices = new_plan.get('prices', {})
        component_map = {
            component.component.type: component
            for component in old_plan.components.all()
        }
        for key, old_component in component_map.items():
            new_amount = new_quotas.get(key, 0)
            if old_component.amount != new_amount:
                old_component.amount = new_amount
                old_component.save(update_fields=['amount'])

            new_price = new_prices.get(key, 0)
            if old_component.price != new_price:
                old_component.price = new_price
                old_component.save(update_fields=['price'])

    def _update_plan_details(self, old_plan, new_plan):
        PLAN_FIELDS = (
            'name', 'description',
            'unit', 'max_amount',
            'product_code', 'article_code',
        )

        for key in PLAN_FIELDS:
            if key in new_plan:
                setattr(old_plan, key, new_plan.get(key))
        old_plan.save()

    def _update_plans(self, offering, new_plans):
        old_plans = offering.plans.all()
        old_ids = set(old_plans.values_list('uuid', flat=True))

        new_map = {plan['uuid']: plan for plan in new_plans if 'uuid' in plan}
        added_plans = [plan for plan in new_plans if 'uuid' not in plan]

        removed_ids = set(old_ids) - set(new_map.keys())
        updated_ids = set(new_map.keys()) & set(old_ids)

        removed_plans = models.Plan.objects.filter(uuid__in=removed_ids).exclude(archived=True)
        updated_plans = {plan.uuid: plan for plan in models.Plan.objects.filter(uuid__in=updated_ids)}

        for plan_uuid, old_plan in updated_plans.items():
            new_plan = new_map[plan_uuid]
            self._update_plan_details(old_plan, new_plan)
            self._update_plan_components(old_plan, new_plan)
            self._update_quotas(old_plan, new_plan)

        if added_plans:
            self._create_plans(offering, added_plans)

        for plan in removed_plans:
            plan.archived = True
            plan.save()

    @transaction.atomic
    def update(self, instance, validated_data):
        """
        Components and plans are specified using nested list serializers with many=True.
        These serializers return empty list even if value is not provided explicitly.
        See also: https://github.com/encode/django-rest-framework/issues/3434
        Consider the case when offering's thumbnail is uploaded, but plans and components are not specified.
        It leads to tricky bug when all components are removed and plans are marked as archived.
        In order to distinguish between case when user asks to remove all plans and
        case when user wants to update only one attribute these we need to check not only
        validated data, but also initial data.
        """
        if 'components' in validated_data:
            components = validated_data.pop('components', [])
            if 'components' in self.initial_data:
                self._update_components(instance, components)
        if 'plans' in validated_data:
            new_plans = validated_data.pop('plans', [])
            if 'plans' in self.initial_data:
                self._update_plans(instance, new_plans)
        limits = validated_data.pop('limits', {})
        if limits:
            self._update_limits(instance, limits)
        offering = super(OfferingUpdateSerializer, self).update(instance, validated_data)
        return offering


class ComponentQuotaSerializer(serializers.ModelSerializer):
    type = serializers.ReadOnlyField(source='component.type')

    class Meta(object):
        model = models.ComponentQuota
        fields = ('type', 'limit', 'usage')


class BaseItemSerializer(core_serializers.AugmentedSerializerMixin,
                         serializers.HyperlinkedModelSerializer,
                         core_serializers.RestrictedSerializerMixin):

    class Meta(object):
        fields = ('offering', 'offering_name', 'offering_uuid',
                  'offering_description', 'offering_thumbnail', 'offering_type',
                  'offering_terms_of_service', 'offering_shared', 'offering_billable',
                  'provider_name', 'provider_uuid',
                  'category_title', 'category_uuid',
                  'plan', 'plan_unit', 'plan_name', 'plan_uuid', 'plan_description',
                  'attributes', 'limits', 'uuid', 'created')
        related_paths = {
            'offering': ('name', 'uuid', 'description', 'type', 'terms_of_service', 'shared', 'billable'),
            'plan': ('unit', 'uuid', 'name', 'description')
        }
        protected_fields = ('offering',)
        extra_kwargs = {
            'offering': {'lookup_field': 'uuid', 'view_name': 'marketplace-offering-detail'},
            'plan': {'lookup_field': 'uuid', 'view_name': 'marketplace-plan-detail'},
        }

    provider_name = serializers.ReadOnlyField(source='offering.customer.name')
    provider_uuid = serializers.ReadOnlyField(source='offering.customer.uuid')
    category_title = serializers.ReadOnlyField(source='offering.category.title')
    category_uuid = serializers.ReadOnlyField(source='offering.category.uuid')
    offering_thumbnail = ProtectedFileField(source='offering.thumbnail', read_only=True)

    def validate_offering(self, offering):
        if not offering.state == models.Offering.States.ACTIVE:
            raise rf_exceptions.ValidationError(_('Offering is not available.'))
        return offering

    def validate(self, attrs):
        offering = attrs.get('offering')
        plan = attrs.get('plan')

        if not offering:
            if not self.instance:
                raise rf_exceptions.ValidationError({
                    'offering': _('This field is required.')
                })
            offering = self.instance.offering

        if plan:
            if plan.offering != offering:
                raise rf_exceptions.ValidationError({
                    'plan': _('This plan is not available for selected offering.')
                })

            validate_plan(plan)

        if offering.options:
            validate_options(offering.options['options'], attrs.get('attributes'))

        limits = attrs.get('limits')
        if limits:
            validate_limits(limits, offering)
        return attrs


class BaseRequestSerializer(BaseItemSerializer):
    type = NaturalChoiceField(
        choices=models.RequestTypeMixin.Types.CHOICES,
        required=False,
        default=models.RequestTypeMixin.Types.CREATE,
    )

    class Meta(BaseItemSerializer.Meta):
        fields = BaseItemSerializer.Meta.fields + ('type',)


class NestedOrderItemSerializer(BaseRequestSerializer):
    class Meta(BaseRequestSerializer.Meta):
        model = models.OrderItem
        fields = BaseRequestSerializer.Meta.fields + (
            'resource_uuid', 'resource_type', 'resource_name',
            'cost', 'state', 'marketplace_resource_uuid', 'error_message',
            'accepting_terms_of_service',
        )

        read_only_fields = ('cost', 'state', 'error_message')
        protected_fields = ('offering', 'plan')

    marketplace_resource_uuid = serializers.ReadOnlyField(source='resource.uuid')
    resource_name = serializers.ReadOnlyField(source='resource.name')
    resource_uuid = serializers.ReadOnlyField(source='resource.backend_uuid')
    resource_type = serializers.ReadOnlyField(source='resource.backend_type')
    state = serializers.ReadOnlyField(source='get_state_display')
    limits = serializers.DictField(child=serializers.IntegerField(), required=False)
    accepting_terms_of_service = serializers.BooleanField(required=False, write_only=True)

    def get_fields(self):
        fields = super(BaseItemSerializer, self).get_fields()
        method = self.context['view'].request.method
        if method == 'GET':
            fields['attributes'] = serializers.ReadOnlyField(source='safe_attributes')
        return fields


class OrderItemDetailsSerializer(NestedOrderItemSerializer):
    class Meta(NestedOrderItemSerializer.Meta):
        fields = NestedOrderItemSerializer.Meta.fields + (
            'order_uuid', 'order_approved_at', 'order_approved_by',
            'created_by_full_name', 'created_by_civil_number',
            'customer_name', 'customer_uuid',
            'project_name', 'project_uuid',
            'old_plan_name', 'new_plan_name',
            'old_plan_uuid', 'new_plan_uuid',
            'old_cost_estimate', 'new_cost_estimate',
            'can_terminate',
        )

    order_uuid = serializers.ReadOnlyField(source='order.uuid')
    order_approved_at = serializers.ReadOnlyField(source='order.approved_at')
    order_approved_by = serializers.ReadOnlyField(source='order.approved_by.full_name')

    created_by_full_name = serializers.ReadOnlyField(source='order.created_by.full_name')
    created_by_civil_number = serializers.ReadOnlyField(source='order.created_by.civil_number')

    customer_name = serializers.ReadOnlyField(source='order.project.customer.name')
    customer_uuid = serializers.ReadOnlyField(source='order.project.customer.uuid')

    project_name = serializers.ReadOnlyField(source='order.project.name')
    project_uuid = serializers.ReadOnlyField(source='order.project.uuid')

    old_plan_name = serializers.ReadOnlyField(source='old_plan.name')
    new_plan_name = serializers.ReadOnlyField(source='plan.name')

    old_plan_uuid = serializers.ReadOnlyField(source='old_plan.uuid')
    new_plan_uuid = serializers.ReadOnlyField(source='plan.uuid')

    old_cost_estimate = serializers.ReadOnlyField(source='resource.cost')
    new_cost_estimate = serializers.ReadOnlyField(source='cost')

    can_terminate = serializers.SerializerMethodField()

    def get_can_terminate(self, order_item):
        if not plugins.manager.can_terminate_order_item(order_item.offering.type):
            return False

        if order_item.state not in (models.OrderItem.States.PENDING, models.OrderItem.States.EXECUTING):
            return False

        return True


class CartItemSerializer(BaseRequestSerializer):
    limits = serializers.DictField(child=serializers.IntegerField(), required=False)
    estimate = serializers.ReadOnlyField(source='cost')
    project = serializers.HyperlinkedRelatedField(
        lookup_field='uuid',
        view_name='project-detail',
        queryset=structure_models.Project.objects.all(),
    )
    project_uuid = serializers.ReadOnlyField(source='project.uuid')
    project_name = serializers.ReadOnlyField(source='project.name')

    class Meta(BaseRequestSerializer.Meta):
        model = models.CartItem
        fields = BaseRequestSerializer.Meta.fields + (
            'estimate', 'project', 'project_name', 'project_uuid'
        )
        protected_fields = BaseRequestSerializer.Meta.protected_fields + ('project',)

    def get_fields(self):
        fields = super(CartItemSerializer, self).get_fields()
        if 'project' in fields:
            fields['project'].queryset = filter_queryset_for_user(
                fields['project'].queryset, self.context['request'].user)
        return fields

    def quotas_validate(self, item, project):
        try:
            with transaction.atomic():
                processor_class = manager.get_processor(item.offering.type, 'create_resource_processor')
                order_params = dict(project=project, created_by=self.context['request'].user)
                order = models.Order(**order_params)
                item_params = get_item_params(item)
                order_item = models.OrderItem(order=order, **item_params)

                if issubclass(processor_class, CreateResourceProcessor):
                    processor = processor_class(order_item)
                    post_data = processor.get_post_data()
                    serializer = processor.get_serializer_class()(data=post_data, context=self.context)
                    serializer.is_valid(raise_exception=True)
                    serializer.save()
                    raise exceptions.TransactionRollback()
        except exceptions.TransactionRollback:
            pass

    @transaction.atomic
    def create(self, validated_data):
        validated_data['user'] = self.context['request'].user
        item = super(CartItemSerializer, self).create(validated_data)
        item.init_cost()
        item.save(update_fields=['cost'])
        self.quotas_validate(item, validated_data['project'])
        return item

    @transaction.atomic
    def update(self, instance, validated_data):
        instance = super(CartItemSerializer, self).update(instance, validated_data)
        instance.init_cost()
        instance.save(update_fields=['cost'])
        return instance


class CartSubmitSerializer(serializers.Serializer):
    project = serializers.HyperlinkedRelatedField(
        queryset=structure_models.Project.objects.all(),
        view_name='project-detail',
        lookup_field='uuid',
        required=True,
    )

    def get_fields(self):
        fields = super(CartSubmitSerializer, self).get_fields()
        project_field = fields['project']
        project_field.queryset = filter_queryset_for_user(
            project_field.queryset, self.context['request'].user)
        return fields

    @transaction.atomic()
    def create(self, validated_data):
        user = self.context['request'].user
        project = validated_data['project']

        items = models.CartItem.objects.filter(user=user, project=project)
        if items.count() == 0:
            raise serializers.ValidationError(_('Shopping cart is empty'))

        order = create_order(project, user, items, self.context['request'])
        items.delete()
        return order


def get_item_params(item):
    return dict(
        offering=item.offering,
        attributes=item.attributes,
        resource=getattr(item, 'resource', None),  # cart item does not have resource
        plan=item.plan,
        old_plan=getattr(item, 'old_plan', None),  # cart item does not have old plan
        limits=item.limits,
        type=item.type,
    )


def create_order(project, user, items, request):
    order_params = dict(project=project, created_by=user)
    order = models.Order.objects.create(**order_params)

    for item in items:
        if item.type in (models.OrderItem.Types.UPDATE, models.OrderItem.Types.TERMINATE) and \
                item.resource and models.OrderItem.objects.filter(
            resource=item.resource,
            state__in=(models.OrderItem.States.PENDING, models.OrderItem.States.EXECUTING)
        ).exists():
            raise rf_exceptions.ValidationError(_('Pending order item for resource already exists.'))

        try:
            params = get_item_params(item)
            order_item = order.add_item(**params)
        except ValidationError as e:
            raise rf_exceptions.ValidationError(e)
        validate_order_item(order_item, request)

    order.init_total_cost()
    order.save()

    if check_availability_of_auto_approving(items, user, project):
        tasks.approve_order(order, user)
    else:
        transaction.on_commit(lambda: tasks.notify_order_approvers.delay(order.uuid))

    return order


class OrderSerializer(structure_serializers.PermissionFieldFilteringMixin,
                      core_serializers.AugmentedSerializerMixin,
                      serializers.HyperlinkedModelSerializer):

    state = serializers.ReadOnlyField(source='get_state_display')
    items = NestedOrderItemSerializer(many=True)

    class Meta(object):
        model = models.Order
        fields = ('url', 'uuid',
                  'created', 'created_by', 'created_by_username', 'created_by_full_name',
                  'approved_by', 'approved_at', 'approved_by_username', 'approved_by_full_name',
                  'project', 'state', 'items', 'total_cost', 'file')
        read_only_fields = ('created_by', 'approved_by', 'approved_at', 'state', 'total_cost')
        protected_fields = ('project', 'items')
        related_paths = {
            'created_by': ('username', 'full_name'),
            'approved_by': ('username', 'full_name'),
        }
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
            'created_by': {'lookup_field': 'uuid', 'view_name': 'user-detail'},
            'approved_by': {'lookup_field': 'uuid', 'view_name': 'user-detail'},
            'project': {'lookup_field': 'uuid', 'view_name': 'project-detail'},
        }

    file = serializers.SerializerMethodField()

    def get_file(self, obj):
        if not obj.has_file():
            return None

        return reverse('marketplace-order-pdf',
                       kwargs={'uuid': obj.uuid},
                       request=self.context['request'])

    @transaction.atomic
    def create(self, validated_data):
        request = self.context['request']
        project = validated_data['project']
        items = [
            models.OrderItem(
                offering=item['offering'],
                plan=item.get('plan'),
                attributes=item.get('attributes', {}),
                limits=item.get('limits', {}),
                type=item.get('type'),
            )
            for item in validated_data['items']
        ]
        return create_order(project, request.user, items, request)

    def get_filtered_field_names(self):
        return 'project',

    def validate_items(self, items):
        for item in items:
            offering = item['offering']

            if offering.shared and offering.terms_of_service and not item.get('accepting_terms_of_service'):
                raise ValidationError(
                    {'items': _('Terms of service for offering \'%s\' have not been accepted.') % offering}
                )

        return items


class CustomerOfferingSerializer(serializers.HyperlinkedModelSerializer):
    offering_set = serializers.HyperlinkedRelatedField(
        many=True,
        view_name='marketplace-offering-detail',
        lookup_field='uuid',
        queryset=models.Offering.objects.all()
    )

    class Meta(object):
        model = structure_models.Customer
        fields = ('offering_set',)


class ResourceSerializer(BaseItemSerializer):
    class Meta(BaseItemSerializer.Meta):
        model = models.Resource
        fields = BaseItemSerializer.Meta.fields + (
            'scope', 'state', 'resource_uuid', 'resource_type',
            'project', 'project_uuid', 'project_name',
            'customer_uuid', 'customer_name',
            'offering_uuid', 'offering_name',
            'backend_metadata', 'is_usage_based', 'name',
            'current_usages',
        )
        read_only_fields = ('backend_metadata', 'scope', 'current_usages')

    state = serializers.ReadOnlyField(source='get_state_display')
    scope = core_serializers.GenericRelatedField()
    resource_uuid = serializers.ReadOnlyField(source='backend_uuid')
    resource_type = serializers.ReadOnlyField(source='backend_type')
    project = serializers.HyperlinkedRelatedField(
        lookup_field='uuid',
        view_name='project-detail',
        read_only=True,
    )
    project_uuid = serializers.ReadOnlyField(source='project.uuid')
    project_name = serializers.ReadOnlyField(source='project.name')
    customer_uuid = serializers.ReadOnlyField(source='project.customer.uuid')
    customer_name = serializers.ReadOnlyField(source='project.customer.name')
    offering_uuid = serializers.ReadOnlyField(source='offering.uuid')
    offering_name = serializers.ReadOnlyField(source='offering.name')
    # If resource is usage-based, frontend would render button to show and report usage
    is_usage_based = serializers.ReadOnlyField(source='offering.is_usage_based')


class ResourceSwitchPlanSerializer(serializers.HyperlinkedModelSerializer):
    class Meta(object):
        model = models.Resource
        fields = ('plan',)

    plan = serializers.HyperlinkedRelatedField(
        view_name='marketplace-plan-detail',
        lookup_field='uuid',
        queryset=models.Plan.objects.all(),
        required=True,
    )

    def validate(self, attrs):
        plan = attrs['plan']
        resource = self.context['view'].get_object()

        if plan.offering != resource.offering:
            raise rf_exceptions.ValidationError({
                'plan': _('Plan is not available for this offering.')
            })

        validate_plan(plan)
        return attrs


class ResourceUpdateLimitsSerializer(serializers.ModelSerializer):
    class Meta(object):
        model = models.Resource
        fields = ('limits',)

    limits = serializers.DictField(child=serializers.IntegerField(min_value=0), required=True)


class BaseComponentSerializer(serializers.Serializer):
    type = serializers.ReadOnlyField(source='component.type')
    name = serializers.ReadOnlyField(source='component.name')
    measured_unit = serializers.ReadOnlyField(source='component.measured_unit')


class CategoryComponentUsageSerializer(core_serializers.RestrictedSerializerMixin,
                                       BaseComponentSerializer,
                                       serializers.ModelSerializer):
    category_title = serializers.ReadOnlyField(source='component.category.title')
    category_uuid = serializers.ReadOnlyField(source='component.category.uuid')
    scope = GenericRelatedField(related_models=(structure_models.Project, structure_models.Customer))

    class Meta(object):
        model = models.CategoryComponentUsage
        fields = ('name', 'type', 'measured_unit', 'category_title', 'category_uuid',
                  'date', 'reported_usage', 'fixed_usage', 'scope')


class BaseComponentUsageSerializer(BaseComponentSerializer, serializers.ModelSerializer):
    class Meta(object):
        model = models.ComponentUsage
        fields = (
            'uuid', 'created', 'description',
            'type', 'name', 'measured_unit', 'usage', 'date',
        )


class ComponentUsageSerializer(BaseComponentUsageSerializer):
    resource_name = serializers.ReadOnlyField(source='resource.name')
    resource_uuid = serializers.ReadOnlyField(source='resource.uuid')

    offering_name = serializers.ReadOnlyField(source='resource.offering.name')
    offering_uuid = serializers.ReadOnlyField(source='resource.offering.uuid')

    project_name = serializers.ReadOnlyField(source='resource.project.name')
    project_uuid = serializers.ReadOnlyField(source='resource.project.uuid')

    customer_name = serializers.ReadOnlyField(source='resource.project.customer.name')
    customer_uuid = serializers.ReadOnlyField(source='resource.project.customer.uuid')

    class Meta(BaseComponentUsageSerializer.Meta):
        fields = BaseComponentUsageSerializer.Meta.fields + (
            'resource_name', 'resource_uuid',
            'offering_name', 'offering_uuid',
            'project_name', 'project_uuid',
            'customer_name', 'customer_uuid',
        )


class ResourcePlanPeriodSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.ResourcePlanPeriod
        fields = ('uuid', 'plan_name', 'plan_uuid', 'start', 'end', 'components')

    plan_name = serializers.ReadOnlyField(source='plan.name')
    plan_uuid = serializers.ReadOnlyField(source='plan.uuid')
    components = BaseComponentUsageSerializer(source='current_components', many=True)


class ImportResourceSerializer(serializers.Serializer):
    backend_id = serializers.CharField()
    project = serializers.SlugRelatedField(queryset=structure_models.Project.objects.all(), slug_field='uuid')
    plan = serializers.SlugRelatedField(queryset=models.Plan.objects.all(), slug_field='uuid')

    def get_fields(self):
        fields = super(ImportResourceSerializer, self).get_fields()

        request = self.context['request']
        user = request.user
        fields['project'].queryset = filter_queryset_for_user(fields['project'].queryset, user)
        return fields


class ServiceProviderSignatureSerializer(serializers.Serializer):
    customer = serializers.SlugRelatedField(queryset=structure_models.Customer.objects.all(), slug_field='uuid')
    data = serializers.CharField()
    dry_run = serializers.BooleanField(default=False, required=False)

    def validate(self, attrs):
        customer = attrs['customer']
        service_provider = getattr(customer, 'serviceprovider', None)
        api_secret_code = service_provider and service_provider.api_secret_code

        if not api_secret_code:
            raise rf_exceptions.ValidationError(_('API secret code is not set.'))

        try:
            data = core_utils.decode_jwt_token(attrs['data'], api_secret_code)
            attrs['data'] = data
            return attrs
        except jwt.exceptions.DecodeError:
            raise rf_exceptions.ValidationError(_('Signature verification failed.'))


class ComponentUsageItemSerializer(serializers.Serializer):
    type = serializers.CharField()
    amount = serializers.IntegerField()
    description = serializers.CharField(required=False, allow_blank=True)


class ComponentUsageCreateSerializer(serializers.Serializer):
    usages = ComponentUsageItemSerializer(many=True)
    plan_period = serializers.SlugRelatedField(queryset=models.ResourcePlanPeriod.objects.all(), slug_field='uuid')

    def validate_plan_period(self, plan_period):
        date = datetime.date.today()
        if plan_period.end and plan_period.end < core_utils.month_start(date):
            raise serializers.ValidationError(_('Billing period is closed.'))
        return plan_period

    def validate(self, attrs):
        attrs = super(ComponentUsageCreateSerializer, self).validate(attrs)
        plan_period = attrs['plan_period']
        resource = plan_period.resource
        offering = resource.plan.offering

        States = models.Resource.States
        if resource.state not in (States.OK, States.UPDATING, States.TERMINATING):
            raise rf_exceptions.ValidationError({
                'resource': _('Resource is not in valid state.')
            })

        valid_components = set(offering.get_usage_components().keys())
        actual_components = {usage['type'] for usage in attrs['usages']}

        missing_components = ', '.join(sorted(valid_components - actual_components))
        invalid_components = ', '.join(sorted(actual_components - valid_components))

        if invalid_components:
            raise rf_exceptions.ValidationError(_('These components are invalid: %s.') % invalid_components)

        if missing_components:
            raise rf_exceptions.ValidationError(_('These components are missing: %s.') % missing_components)

        return attrs

    def save(self):
        plan_period = self.validated_data['plan_period']
        resource = plan_period.resource
        components = resource.plan.offering.get_usage_components()
        now = timezone.now()
        billing_period = core_utils.month_start(now)

        for usage in self.validated_data['usages']:
            amount = usage['amount']
            description = usage.get('description', '')
            component = components[usage['type']]
            component.validate_amount(resource, amount, now)

            models.ComponentUsage.objects.update_or_create(
                resource=resource,
                component=component,
                plan_period=plan_period,
                billing_period=billing_period,
                defaults={'usage': amount, 'date': now, 'description': description},
            )


class OfferingFileSerializer(ProtectedMediaSerializerMixin,
                             core_serializers.RestrictedSerializerMixin,
                             core_serializers.AugmentedSerializerMixin,
                             serializers.HyperlinkedModelSerializer):

    class Meta(object):
        model = models.OfferingFile
        fields = ('url', 'uuid', 'name', 'offering', 'created', 'file',)
        extra_kwargs = dict(
            url={'lookup_field': 'uuid'},
            offering={'lookup_field': 'uuid', 'view_name': 'marketplace-offering-detail'},
        )


def validate_plan(plan):
    """"
    Ensure that maximum amount of resources with current plan is not reached yet.
    """
    if not plan.is_active:
        raise rf_exceptions.ValidationError({
            'plan': _('Plan is not available because limit has been reached.')
        })


def get_is_service_provider(serializer, scope):
    customer = structure_permissions._get_customer(scope)
    return models.ServiceProvider.objects.filter(customer=customer).exists()


def add_service_provider(sender, fields, **kwargs):
    fields['is_service_provider'] = serializers.SerializerMethodField()
    setattr(sender, 'get_is_service_provider', get_is_service_provider)


def get_marketplace_offering_uuid(serializer, scope):
    try:
        return models.Resource.objects.get(scope=scope).offering.uuid
    except ObjectDoesNotExist:
        return


def get_marketplace_offering_name(serializer, scope):
    try:
        return models.Resource.objects.get(scope=scope).offering.name
    except ObjectDoesNotExist:
        return


def get_marketplace_category_uuid(serializer, scope):
    try:
        return models.Resource.objects.get(scope=scope).offering.category.uuid
    except ObjectDoesNotExist:
        return


def get_marketplace_category_name(serializer, scope):
    try:
        return models.Resource.objects.get(scope=scope).offering.category.title
    except ObjectDoesNotExist:
        return


def get_marketplace_resource_uuid(serializer, scope):
    try:
        return models.Resource.objects.get(scope=scope).uuid
    except ObjectDoesNotExist:
        return


def get_marketplace_resource_state(serializer, scope):
    try:
        return models.Resource.objects.get(scope=scope).get_state_display()
    except ObjectDoesNotExist:
        return


def get_is_usage_based(serializer, scope):
    try:
        return models.Resource.objects.get(scope=scope).offering.is_usage_based
    except ObjectDoesNotExist:
        return


def add_marketplace_offering(sender, fields, **kwargs):
    fields['marketplace_offering_uuid'] = serializers.SerializerMethodField()
    setattr(sender, 'get_marketplace_offering_uuid', get_marketplace_offering_uuid)

    fields['marketplace_offering_name'] = serializers.SerializerMethodField()
    setattr(sender, 'get_marketplace_offering_name', get_marketplace_offering_name)

    fields['marketplace_category_uuid'] = serializers.SerializerMethodField()
    setattr(sender, 'get_marketplace_category_uuid', get_marketplace_category_uuid)

    fields['marketplace_category_name'] = serializers.SerializerMethodField()
    setattr(sender, 'get_marketplace_category_name', get_marketplace_category_name)

    fields['marketplace_resource_uuid'] = serializers.SerializerMethodField()
    setattr(sender, 'get_marketplace_resource_uuid', get_marketplace_resource_uuid)

    fields['marketplace_resource_state'] = serializers.SerializerMethodField()
    setattr(sender, 'get_marketplace_resource_state', get_marketplace_resource_state)

    fields['is_usage_based'] = serializers.SerializerMethodField()
    setattr(sender, 'get_is_usage_based', get_is_usage_based)


class ResourceTerminateSerializer(serializers.Serializer):
    attributes = serializers.JSONField(label=_('Termination attributes'), required=False)


core_signals.pre_serializer_fields.connect(
    sender=structure_serializers.CustomerSerializer,
    receiver=add_service_provider,
)

core_signals.pre_serializer_fields.connect(
    sender=support_serializers.OfferingSerializer,
    receiver=add_marketplace_offering,
)

for resource_serializer in SupportedServices.get_resource_serializers():
    core_signals.pre_serializer_fields.connect(
        sender=resource_serializer,
        receiver=add_marketplace_offering,
    )



from django.apps import AppConfig
from django.db.models import signals


class SupportConfig(AppConfig):
    name = 'waldur_mastermind.support'
    verbose_name = 'HelpDesk'

    def ready(self):
        from waldur_core.quotas import fields as quota_fields
        from waldur_core.structure import models as structure_models

        from . import handlers

        Issue = self.get_model('Issue')
        Attachment = self.get_model('Attachment')
        Offering = self.get_model('Offering')
        Comment = self.get_model('Comment')

        structure_models.Project.add_quota_field(
            name='nc_offering_count',
            quota_field=quota_fields.CounterQuotaField(
                target_models=[Offering],
                path_to_scope='project',
            )
        )
        structure_models.Customer.add_quota_field(
            name='nc_offering_count',
            quota_field=quota_fields.CounterQuotaField(
                target_models=[Offering],
                path_to_scope='project.customer',
            )
        )

        signals.post_save.connect(
            handlers.log_issue_save,
            sender=Issue,
            dispatch_uid='waldur_mastermind.support.handlers.log_issue_save',
        )

        signals.post_delete.connect(
            handlers.log_issue_delete,
            sender=Issue,
            dispatch_uid='waldur_mastermind.support.handlers.log_issue_delete',
        )

        signals.post_save.connect(
            handlers.log_attachment_save,
            sender=Attachment,
            dispatch_uid='waldur_mastermind.support.handlers.log_attachment_save',
        )

        signals.post_delete.connect(
            handlers.log_attachment_delete,
            sender=Attachment,
            dispatch_uid='waldur_mastermind.support.handlers.log_attachment_delete',
        )

        signals.post_save.connect(
            handlers.log_offering_created,
            sender=Offering,
            dispatch_uid='waldur_mastermind.support.handlers.log_offering_created',
        )

        signals.pre_delete.connect(
            handlers.log_offering_deleted,
            sender=Offering,
            dispatch_uid='waldur_mastermind.support.handlers.log_offering_deleted',
        )

        signals.post_save.connect(
            handlers.log_offering_state_changed,
            sender=Offering,
            dispatch_uid='waldur_mastermind.support.handlers.log_offering_state_changed',
        )

        signals.post_save.connect(
            handlers.send_comment_added_notification,
            sender=Comment,
            dispatch_uid='waldur_mastermind.support.handlers.send_comment_added_notification'
        )

        signals.post_save.connect(
            handlers.send_issue_updated_notification,
            sender=Issue,
            dispatch_uid='waldur_mastermind.support.handlers.send_issue_updated_notification'
        )



# -*- coding: utf-8 -*-
# Generated by Django 1.11.18 on 2019-02-11 10:56
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('invoices', '0027_migrate_items'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='offeringitem',
            name='invoice',
        ),
        migrations.RemoveField(
            model_name='offeringitem',
            name='offering',
        ),
        migrations.RemoveField(
            model_name='offeringitem',
            name='project',
        ),
        migrations.RemoveField(
            model_name='openstackitem',
            name='invoice',
        ),
        migrations.RemoveField(
            model_name='openstackitem',
            name='package',
        ),
        migrations.RemoveField(
            model_name='openstackitem',
            name='project',
        ),
        migrations.DeleteModel(
            name='OfferingItem',
        ),
        migrations.DeleteModel(
            name='OpenStackItem',
        ),
    ]



import logging
import sys
import ssl

from django.utils import six, timezone
from django.utils.functional import cached_property
import pyVim.task
import pyVim.connect
from pyVmomi import vim
from six.moves.urllib import parse as urlparse

from waldur_core.structure import ServiceBackend, ServiceBackendError, log_backend_action
from waldur_core.structure.utils import update_pulled_fields
from waldur_mastermind.common.utils import parse_datetime
from waldur_vmware.client import VMwareClient
from waldur_vmware.exceptions import VMwareError
from waldur_vmware.utils import is_basic_mode

from . import models, signals

logger = logging.getLogger(__name__)


class VMwareBackendError(ServiceBackendError):
    pass


def reraise(exc):
    """
    Reraise VMwareBackendError while maintaining traceback.
    """
    six.reraise(VMwareBackendError, exc, sys.exc_info()[2])


class VMwareBackend(ServiceBackend):
    def __init__(self, settings):
        """
        :type settings: :class:`waldur_core.structure.models.ServiceSettings`
        """
        self.settings = settings

    @cached_property
    def host(self):
        return self.settings.backend_url.split('https://')[-1].split('http://')[-1].strip('/')

    @cached_property
    def client(self):
        """
        Construct VMware REST API client using credentials specified in the service settings.
        """
        client = VMwareClient(self.host, verify_ssl=False)
        client.login(self.settings.username, self.settings.password)
        return client

    @cached_property
    def soap_client(self):
        """
        Construct VMware SOAP API client using credentials specified in the service settings.
        """
        context = ssl.SSLContext(ssl.PROTOCOL_TLSv1_2)
        context.verify_mode = ssl.CERT_NONE
        return pyVim.connect.SmartConnect(
            host=self.host,
            user=self.settings.username,
            pwd=self.settings.password,
            port=443,
            sslContext=context
        )

    def ping(self, raise_exception=False):
        """
        Check if backend is ok.
        """
        try:
            self.client.list_vms()
        except VMwareError as e:
            if raise_exception:
                reraise(e)
            return False
        else:
            return True

    def pull_service_properties(self):
        self.pull_folders()
        self.pull_templates()
        self.pull_clusters()
        self.pull_networks()
        self.pull_datastores()

    def pull_templates(self):
        """
        Pull VMware templates for virtual machine provisioning from content library
        using VMware REST API to the local database.
        """
        try:
            backend_templates = self.client.list_all_templates()
        except VMwareError as e:
            reraise(e)
            return

        if is_basic_mode():
            # If basic mode is enabled, we should filter out templates which have more than 1 NIC
            backend_templates = [
                template for template in backend_templates
                if len(template['template']['nics']) == 1
            ]

        backend_templates_map = {
            item['library_item']['id']: item
            for item in backend_templates
        }

        frontend_templates_map = {
            p.backend_id: p
            for p in models.Template.objects.filter(settings=self.settings)
        }

        stale_ids = set(frontend_templates_map.keys()) - set(backend_templates_map.keys())
        new_ids = set(backend_templates_map.keys()) - set(frontend_templates_map.keys())
        common_ids = set(backend_templates_map.keys()) & set(frontend_templates_map.keys())

        for library_item_id in new_ids:
            template = self._backend_template_to_template(backend_templates_map[library_item_id])
            template.save()

        for library_item_id in common_ids:
            backend_template = self._backend_template_to_template(
                backend_templates_map[library_item_id])
            frontend_template = frontend_templates_map[library_item_id]
            fields = (
                'cores',
                'cores_per_socket',
                'ram',
                'disk',
                'guest_os',
                'modified',
                'description'
            )
            update_pulled_fields(frontend_template, backend_template, fields)

        models.Template.objects.filter(settings=self.settings, backend_id__in=stale_ids).delete()

    def _backend_template_to_template(self, backend_template):
        library_item = backend_template['library_item']
        template = backend_template['template']
        total_disk = self._get_total_disk(template['disks'])
        return models.Template(
            settings=self.settings,
            backend_id=library_item['id'],
            name=library_item['name'],
            description=library_item['description'],
            created=parse_datetime(library_item['creation_time']),
            modified=parse_datetime(library_item['last_modified_time']),
            cores=template['cpu']['count'],
            cores_per_socket=template['cpu']['cores_per_socket'],
            ram=template['memory']['size_MiB'],
            disk=total_disk,
            guest_os=template['guest_OS'],
        )

    def _get_total_disk(self, backend_disks):
        # Convert disk size from bytes to MiB
        return sum([disk['value']['capacity'] / 1024 / 1024 for disk in backend_disks])

    @log_backend_action()
    def pull_virtual_machine(self, vm, update_fields=None):
        """
        Pull virtual machine from REST API and update its information in local database.

        :param vm: Virtual machine database object.
        :type vm: :class:`waldur_vmware.models.VirtualMachine`
        :param update_fields: iterable of fields to be updated
        """
        import_time = timezone.now()
        imported_vm = self.import_virtual_machine(vm.backend_id, save=False)

        vm.refresh_from_db()
        if vm.modified < import_time:
            if not update_fields:
                update_fields = models.VirtualMachine.get_backend_fields()

            update_pulled_fields(vm, imported_vm, update_fields)

    def import_virtual_machine(self, backend_id, save=True, service_project_link=None):
        """
        Import virtual machine by its ID.

        :param backend_id: Virtual machine identifier
        :type backend_id: str
        :param save: Save object in the database
        :type save: bool
        :param service_project_link: Optional service project link model object
        :rtype: :class:`waldur_vmware.models.VirtualMachine`
        """
        try:
            backend_vm = self.client.get_vm(backend_id)
        except VMwareError as e:
            reraise(e)
            return

        tools_installed = self.get_vm_tools_installed(backend_id)
        tools_state = self.get_vm_tools_state(backend_id)

        vm = self._backend_vm_to_vm(backend_vm, tools_installed, tools_state, backend_id)
        if service_project_link is not None:
            vm.service_project_link = service_project_link
        if save:
            vm.save()

        return vm

    def _backend_vm_to_vm(self, backend_vm, tools_installed, tools_state, backend_id):
        """
        Build database model object for virtual machine from REST API spec.

        :param backend_vm: virtual machine specification
        :type backend_vm: dict
        :param tools_installed: whether VMware tools installed or not
        :type tools_installed: bool
        :param tools_state: Status of VMware Tools.
        :type tools_state: str
        :param backend_id: Virtual machine identifier
        :type backend_id: str
        :rtype: :class:`waldur_vmware.models.VirtualMachine`
        """
        return models.VirtualMachine(
            backend_id=backend_id,
            name=backend_vm['name'],
            state=models.VirtualMachine.States.OK,
            runtime_state=backend_vm['power_state'],
            cores=backend_vm['cpu']['count'],
            cores_per_socket=backend_vm['cpu']['cores_per_socket'],
            ram=backend_vm['memory']['size_MiB'],
            disk=self._get_total_disk(backend_vm['disks']),
            tools_installed=tools_installed,
            tools_state=tools_state,
        )

    def pull_clusters(self):
        try:
            backend_clusters = self.client.list_clusters()
        except VMwareError as e:
            reraise(e)
            return

        backend_clusters_map = {
            item['cluster']: item
            for item in backend_clusters
        }

        frontend_clusters_map = {
            p.backend_id: p
            for p in models.Cluster.objects.filter(settings=self.settings)
        }

        stale_ids = set(frontend_clusters_map.keys()) - set(backend_clusters_map.keys())
        new_ids = set(backend_clusters_map.keys()) - set(frontend_clusters_map.keys())
        common_ids = set(backend_clusters_map.keys()) & set(frontend_clusters_map.keys())

        for item_id in common_ids:
            backend_item = backend_clusters_map[item_id]
            frontend_item = frontend_clusters_map[item_id]
            if frontend_item.name != backend_item['name']:
                frontend_item.name = backend_item['name']
                frontend_item.save(update_fields=['name'])

        for item_id in new_ids:
            item = backend_clusters_map[item_id]
            models.Cluster.objects.create(
                settings=self.settings,
                backend_id=item_id,
                name=item['name'],
            )

        models.Cluster.objects.filter(settings=self.settings, backend_id__in=stale_ids).delete()

    def pull_networks(self):
        try:
            backend_networks = self.client.list_networks()
        except VMwareError as e:
            reraise(e)
            return

        backend_networks_map = {
            item['network']: item
            for item in backend_networks
        }

        frontend_networks_map = {
            p.backend_id: p
            for p in models.Network.objects.filter(settings=self.settings)
        }

        stale_ids = set(frontend_networks_map.keys()) - set(backend_networks_map.keys())
        new_ids = set(backend_networks_map.keys()) - set(frontend_networks_map.keys())
        common_ids = set(frontend_networks_map.keys()) & set(backend_networks_map.keys())

        for item_id in common_ids:
            backend_item = backend_networks_map[item_id]
            frontend_item = frontend_networks_map[item_id]
            if frontend_item.name != backend_item['name']:
                frontend_item.name = backend_item['name']
                frontend_item.save(update_fields=['name'])

        for item_id in new_ids:
            item = backend_networks_map[item_id]
            models.Network.objects.create(
                settings=self.settings,
                backend_id=item_id,
                name=item['name'],
                type=item['type'],
            )

        models.Network.objects.filter(settings=self.settings, backend_id__in=stale_ids).delete()

    def pull_datastores(self):
        try:
            backend_datastores = self.client.list_datastores()
        except VMwareError as e:
            reraise(e)
            return

        backend_datastores_map = {
            item['datastore']: item
            for item in backend_datastores
        }

        frontend_datastores_map = {
            p.backend_id: p
            for p in models.Datastore.objects.filter(settings=self.settings)
        }

        stale_ids = set(frontend_datastores_map.keys()) - set(backend_datastores_map.keys())
        new_ids = set(backend_datastores_map.keys()) - set(frontend_datastores_map.keys())
        common_ids = set(backend_datastores_map.keys()) & set(frontend_datastores_map.keys())

        for item_id in new_ids:
            datastore = self._backend_datastore_to_datastore(backend_datastores_map[item_id])
            datastore.save()

        for item_id in common_ids:
            backend_datastore = self._backend_datastore_to_datastore(backend_datastores_map[item_id])
            frontend_datastore = frontend_datastores_map[item_id]
            fields = ('name', 'capacity', 'free_space')
            update_pulled_fields(frontend_datastore, backend_datastore, fields)

        models.Datastore.objects.filter(settings=self.settings, backend_id__in=stale_ids).delete()

    def _backend_datastore_to_datastore(self, backend_datastore):
        capacity = backend_datastore.get('capacity')
        # Convert from bytes to MB
        if capacity:
            capacity /= 1024 * 1024

        free_space = backend_datastore.get('free_space')
        # Convert from bytes to MB
        if free_space:
            free_space /= 1024 * 1024

        return models.Datastore(
            settings=self.settings,
            backend_id=backend_datastore['datastore'],
            name=backend_datastore['name'],
            type=backend_datastore['type'],
            capacity=capacity,
            free_space=free_space,
        )

    def get_vm_folders(self):
        try:
            return self.client.list_folders(folder_type='VIRTUAL_MACHINE')
        except VMwareError as e:
            reraise(e)

    def get_default_vm_folder(self):
        """
        Currently VM folder is required for VM provisioning either from template or from scratch.
        Therefore when folder is not specified for VM, we should use first available folder.
        Please note that it is assumed that there's only one datacenter in this case.
        :return: Virtual machine folder identifier.
        :rtype: str
        """
        return self.get_vm_folders()[0]['folder']

    def get_default_resource_pool(self):
        """
        Currently resource pool is required for VM provisioning from scratch if cluster is not specified.
        Therefore we should use first available resource pool.
        Please note that it is assumed that there's only one datacenter in this case.
        :return: Resource pool identifier.
        :rtype: str
        """
        try:
            return self.client.list_resource_pools()[0]['resource_pool']
        except VMwareError as e:
            reraise(e)

    def get_default_datastore(self):
        """
        Currently datastore is required for VM provisioning either from template or from scratch.
        Therefore when datastore is not specified for VM, we should use first available datastore.
        Please note that it is assumed that there's only one datacenter in this case.
        :return: Datastore identifier.
        :rtype: str
        """
        try:
            return self.client.list_datastores()[0]['datastore']
        except VMwareError as e:
            reraise(e)

    def pull_folders(self):
        backend_folders = self.get_vm_folders()

        backend_folders_map = {
            item['folder']: item
            for item in backend_folders
        }

        frontend_folders_map = {
            p.backend_id: p
            for p in models.Folder.objects.filter(settings=self.settings)
        }

        stale_ids = set(frontend_folders_map.keys()) - set(backend_folders_map.keys())
        new_ids = set(backend_folders_map.keys()) - set(frontend_folders_map.keys())
        common_ids = set(backend_folders_map.keys()) & set(frontend_folders_map.keys())

        for item_id in common_ids:
            backend_item = backend_folders_map[item_id]
            frontend_item = frontend_folders_map[item_id]
            if frontend_item.name != backend_item['name']:
                frontend_item.name = backend_item['name']
                frontend_item.save(update_fields=['name'])

        for item_id in new_ids:
            item = backend_folders_map[item_id]
            models.Folder.objects.create(
                settings=self.settings,
                backend_id=item_id,
                name=item['name'],
            )

        models.Folder.objects.filter(settings=self.settings, backend_id__in=stale_ids).delete()

    def create_virtual_machine(self, vm):
        """
        Creates a virtual machine.

        :param vm: Virtual machine to be created
        :type vm: :class:`waldur_vmware.models.VirtualMachine`
        """
        if vm.template:
            backend_id = self.create_virtual_machine_from_template(vm)
        else:
            backend_id = self.create_virtual_machine_from_scratch(vm)

        try:
            backend_vm = self.client.get_vm(backend_id)
        except VMwareError as e:
            reraise(e)
            return

        vm.backend_id = backend_id
        vm.runtime_state = backend_vm['power_state']
        vm.save(update_fields=['backend_id', 'runtime_state'])

        for disk in backend_vm['disks']:
            disk = self._backend_disk_to_disk(disk['value'], disk['key'])
            disk.vm = vm
            disk.service_project_link = vm.service_project_link
            disk.save()

        # If virtual machine is not deployed from template, it does not have any networks.
        # Therefore we should create network interfaces manually according to VM spec.
        if not vm.template:
            for network in vm.networks.all():
                try:
                    self.client.create_nic(vm.backend_id, network.backend_id)
                except VMwareError as e:
                    reraise(e)

        signals.vm_created.send(self.__class__, vm=vm)
        return vm

    def _get_vm_placement(self, vm):
        placement = {}

        if vm.folder:
            placement['folder'] = vm.folder.backend_id
        else:
            logger.warning('Folder is not specified for VM with ID: %s. '
                           'Trying to assign default folder.', vm.id)
            placement['folder'] = self.get_default_vm_folder()

        if vm.cluster:
            placement['cluster'] = vm.cluster.backend_id
        else:
            logger.warning('Cluster is not specified for VM with ID: %s. '
                           'Trying to assign default resource pool.', vm.id)
            placement['resource_pool'] = self.get_default_resource_pool()

        return placement

    def _get_template_nics(self, template):
        """
        Fetch list of NIC IDs assigned to virtual machine template.

        :param template: Virtual machine template.
        :type template: :class:`waldur_vmware.models.Template`
        :rtype: list[str]
        """

        try:
            backend_template = self.client.get_template_library_item(template.backend_id)
        except VMwareError as e:
            reraise(e)
        else:
            return [nic['key'] for nic in backend_template['nics']]

    def _get_vm_nics(self, vm):
        """
        Serialize map of Ethernet network adapters for virtual machine template deployment.

        :param vm: Virtual machine to be created.
        :type vm: :class:`waldur_vmware.models.VirtualMachine`
        :return: list[dict]
        """

        nics = self._get_template_nics(vm.template)
        networks = list(vm.networks.all())

        if is_basic_mode():
            if len(networks) != 1:
                logger.warning('Skipping network assignment because VM does not have '
                               'exactly one network in basic mode. VM ID: %s', vm.id)
                return
            elif len(nics) != 1:
                logger.warning('Skipping network assignment because related template does '
                               'not have exactly one NIC in basic mode. VM ID: %s', vm.id)

        if len(networks) != len(nics):
            logger.warning('It is not safe to update network assignment when '
                           'number of interfaces and networks do not match. VM ID: %s', vm.id)

        return [
            {
                'key': nic,
                'value': {
                    'network': network.backend_id
                }
            }
            for (nic, network) in zip(nics, networks)
        ]

    def create_virtual_machine_from_template(self, vm):
        spec = {
            'name': vm.name,
            'description': vm.description,
            'hardware_customization': {
                'cpu_update': {
                    'num_cpus': vm.cores,
                    'num_cores_per_socket': vm.cores_per_socket,
                },
                'memory_update': {
                    'memory': vm.ram,
                },
            },
            'placement': self._get_vm_placement(vm),
        }

        if vm.datastore:
            spec['disk_storage'] = {'datastore': vm.datastore.backend_id}
            spec['vm_home_storage'] = {'datastore': vm.datastore.backend_id}

        nics = self._get_vm_nics(vm)
        if nics:
            spec['hardware_customization']['nics'] = nics

        try:
            return self.client.deploy_vm_from_template(vm.template.backend_id, spec)
        except VMwareError as e:
            reraise(e)

    def create_virtual_machine_from_scratch(self, vm):
        spec = {
            'name': vm.name,
            'guest_OS': vm.guest_os,
            'cpu': {
                'count': vm.cores,
                'cores_per_socket': vm.cores_per_socket,
                'hot_add_enabled': True,
                'hot_remove_enabled': True
            },
            'memory': {
                'size_MiB': vm.ram,
                'hot_add_enabled': True,
            },
            'placement': self._get_vm_placement(vm),
        }

        if vm.datastore:
            spec['placement']['datastore'] = vm.datastore.backend_id
        else:
            spec['placement']['datastore'] = self.get_default_datastore()

        try:
            return self.client.create_vm(spec)
        except VMwareError as e:
            reraise(e)

    def delete_virtual_machine(self, vm):
        """
        Deletes a virtual machine.

        :param vm: Virtual machine to be deleted
        :type vm: :class:`waldur_vmware.models.VirtualMachine`
        """
        try:
            self.client.delete_vm(vm.backend_id)
        except VMwareError as e:
            reraise(e)

    def start_virtual_machine(self, vm):
        """
        Powers on a powered-off or suspended virtual machine.

        :param vm: Virtual machine to be started
        :type vm: :class:`waldur_vmware.models.VirtualMachine`
        """
        try:
            self.client.start_vm(vm.backend_id)
        except VMwareError as e:
            reraise(e)

    def stop_virtual_machine(self, vm):
        """
        Powers off a powered-on or suspended virtual machine.

        :param vm: Virtual machine to be stopped
        :type vm: :class:`waldur_vmware.models.VirtualMachine`
        """
        try:
            self.client.stop_vm(vm.backend_id)
        except VMwareError as e:
            reraise(e)

    def reset_virtual_machine(self, vm):
        """
        Resets a powered-on virtual machine.

        :param vm: Virtual machine.
        :type vm: :class:`waldur_vmware.models.VirtualMachine`
        """
        try:
            self.client.reset_vm(vm.backend_id)
        except VMwareError as e:
            reraise(e)

    def suspend_virtual_machine(self, vm):
        """
        Suspends a powered-on virtual machine.

        :param vm: Virtual machine.
        :type vm: :class:`waldur_vmware.models.VirtualMachine`
        """
        try:
            self.client.suspend_vm(vm.backend_id)
        except VMwareError as e:
            reraise(e)

    def shutdown_guest(self, vm):
        """
        Issues a request to the guest operating system asking
        it to perform a clean shutdown of all services.

        :param vm: Virtual machine.
        :type vm: :class:`waldur_vmware.models.VirtualMachine`
        """
        try:
            self.client.shutdown_guest(vm.backend_id)
        except VMwareError as e:
            reraise(e)

    def reboot_guest(self, vm):
        """
        Issues a request to the guest operating system asking it to perform a reboot.

        :param vm: Virtual machine.
        :type vm: :class:`waldur_vmware.models.VirtualMachine`
        """
        try:
            self.client.reboot_guest(vm.backend_id)
        except VMwareError as e:
            reraise(e)

    def is_virtual_machine_shutted_down(self, vm):
        try:
            guest_power = self.client.get_guest_power(vm.backend_id)
        except VMwareError as e:
            reraise(e)
        else:
            return guest_power['state'] == models.VirtualMachine.GuestPowerStates.NOT_RUNNING

    def is_virtual_machine_tools_running(self, vm):
        """
        Check VMware tools status and update cache only if its running.
        If VMware tools are not running, state is not updated.
        It is needed in order to skip extra database updates.
        Otherwise VMware tools state in database would be updated
        from RUNNING to NOT RUNNING twice when optimistic update is used.
        """
        tools_state = self.get_vm_tools_state(vm.backend_id)
        result = tools_state == models.VirtualMachine.ToolsStates.RUNNING
        if result:
            vm.tools_state = tools_state
            vm.save(update_fields=['tools_state'])
        self.pull_virtual_machine_runtime_state(vm)
        return result

    def pull_virtual_machine_runtime_state(self, vm):
        try:
            backend_vm = self.client.get_vm(vm.backend_id)
        except VMwareError as e:
            reraise(e)
        else:
            backend_power_state = backend_vm['power_state']
            if backend_power_state != vm.runtime_state:
                vm.runtime_state = backend_power_state
                vm.save(update_fields=['runtime_state'])

    def is_virtual_machine_tools_not_running(self, vm):
        tools_state = self.get_vm_tools_state(vm.backend_id)
        result = tools_state == models.VirtualMachine.ToolsStates.NOT_RUNNING
        if result:
            vm.tools_state = tools_state
            vm.save(update_fields=['tools_state'])
        return result

    def update_virtual_machine(self, vm):
        """
        Updates CPU and RAM of virtual machine.
        """
        self.update_cpu(vm)
        self.update_memory(vm)
        signals.vm_updated.send(self.__class__, vm=vm)

    def update_cpu(self, vm):
        """
        Updates CPU of virtual machine.

        :param vm: Virtual machine.
        :type vm: :class:`waldur_vmware.models.VirtualMachine`
        """
        try:
            cpu_spec = self.client.get_cpu(vm.backend_id)
            if cpu_spec['cores_per_socket'] != vm.cores_per_socket or cpu_spec['count'] != vm.cores:
                self.client.update_cpu(vm.backend_id, {
                    'cores_per_socket': vm.cores_per_socket,
                    'count': vm.cores,
                })
        except VMwareError as e:
            reraise(e)

    def update_memory(self, vm):
        """
        Updates RAM of virtual machine.

        :param vm: Virtual machine.
        :type vm: :class:`waldur_vmware.models.VirtualMachine`
        """
        try:
            memory_spec = self.client.get_memory(vm.backend_id)
            if memory_spec['size_MiB'] != vm.ram:
                self.client.update_memory(vm.backend_id, {
                    'size_MiB': vm.ram
                })
        except VMwareError as e:
            reraise(e)

    def create_port(self, port):
        """
        Creates an Ethernet port for given VM and network.

        :param port: Port to be created
        :type port: :class:`waldur_vmware.models.Port`
        """
        try:
            backend_id = self.client.create_nic(port.vm.backend_id, port.network.backend_id)
        except VMwareError as e:
            reraise(e)
        else:
            port.backend_id = backend_id
            port.save(update_fields=['backend_id'])
            return port

    def delete_port(self, port):
        """
        Deletes an Ethernet port.

        :param port: Port to be deleted.
        :type port: :class:`waldur_vmware.models.Port`
        """
        try:
            self.client.delete_nic(port.vm.backend_id, port.backend_id)
        except VMwareError as e:
            reraise(e)

    @log_backend_action()
    def pull_port(self, port, update_fields=None):
        """
        Pull Ethernet port from REST API and update its information in local database.

        :param port: Port to be updated.
        :type port: :class:`waldur_vmware.models.Port`
        :param update_fields: iterable of fields to be updated
        :return: None
        """
        import_time = timezone.now()
        imported_port = self.import_port(port.vm.backend_id, port.backend_id, save=False)

        port.refresh_from_db()
        if port.modified < import_time:
            if not update_fields:
                update_fields = models.Port.get_backend_fields()

            update_pulled_fields(port, imported_port, update_fields)

    def import_port(self, backend_vm_id, backend_port_id, save=True, service_project_link=None):
        """
        Import Ethernet port by its ID.

        :param backend_vm_id: Virtual machine identifier
        :type backend_vm_id: str
        :param backend_port_id: Ethernet port identifier
        :type backend_port_id: str
        :param save: Save object in the database
        :type save: bool
        :param service_project_link: Service project link model object
        :rtype: :class:`waldur_vmware.models.Disk`
        """
        try:
            backend_port = self.client.get_nic(backend_vm_id, backend_port_id)
            backend_port['nic'] = backend_port_id
        except VMwareError as e:
            reraise(e)
            return

        port = self._backend_port_to_port(backend_port)
        if service_project_link is not None:
            port.service_project_link = service_project_link
        if save:
            port.save()

        return port

    def _backend_port_to_port(self, backend_port):
        """
        Build database model object for Ethernet port from REST API spec.

        :param backend_port: Ethernet port specification
        :type backend_port: dict
        :rtype: :class:`waldur_vmware.models.Port`
        """
        return models.Port(
            backend_id=backend_port['nic'],
            name=backend_port['label'],
            # MAC address is optional
            mac_address=backend_port.get('mac_address'),
            state=models.Port.States.OK,
            runtime_state=backend_port['state'],
        )

    def pull_vm_ports(self, vm):
        try:
            backend_ports = self.client.list_nics(vm.backend_id)
        except VMwareError as e:
            reraise(e)
            return

        backend_ports_map = {
            item['nic']: item
            for item in backend_ports
        }

        frontend_ports_map = {
            p.backend_id: p
            for p in models.Port.objects.filter(vm=vm)
        }

        networks_map = {
            p.backend_id: p
            for p in models.Network.objects.filter(settings=vm.service_settings)
        }

        stale_ids = set(frontend_ports_map.keys()) - set(backend_ports_map.keys())
        new_ids = set(backend_ports_map.keys()) - set(frontend_ports_map.keys())
        common_ids = set(backend_ports_map.keys()) & set(frontend_ports_map.keys())

        for item_id in new_ids:
            backend_port = backend_ports_map[item_id]
            port = self._backend_port_to_port(backend_port)
            port.service_project_link = vm.service_project_link
            network_id = backend_port['backing']['network']
            port.network = networks_map.get(network_id)
            port.vm = vm
            port.save()

        for item_id in common_ids:
            backend_port = self._backend_port_to_port(backend_ports_map[item_id])
            frontend_port = frontend_ports_map[item_id]
            fields = ('mac_address', 'runtime_state')
            update_pulled_fields(frontend_port, backend_port, fields)

        models.Port.objects.filter(vm=vm, backend_id__in=stale_ids).delete()

    def create_disk(self, disk):
        """
        Creates a virtual disk.

        :param disk: Virtual disk to be created
        :type disk: :class:`waldur_vmware.models.Disk`
        """
        spec = {
            'new_vmdk': {
                # Convert from mebibytes to bytes because VMDK is specified in bytes
                'capacity': 1024 * 1024 * disk.size,
            }
        }

        try:
            backend_id = self.client.create_disk(disk.vm.backend_id, spec)
        except VMwareError as e:
            reraise(e)
        else:
            disk.backend_id = backend_id
            disk.save(update_fields=['backend_id'])
            signals.vm_updated.send(self.__class__, vm=disk.vm)
            return disk

    def delete_disk(self, disk, delete_vmdk=True):
        """
        Deletes a virtual disk.

        :param disk: Virtual disk to be deleted
        :type disk: :class:`waldur_vmware.models.Disk`
        :param delete_vmdk: Delete backing VMDK file.
        """
        backend_disk = self.get_backend_disk(disk)

        try:
            self.client.delete_disk(disk.vm.backend_id, disk.backend_id)
        except VMwareError as e:
            reraise(e)

        if delete_vmdk:
            vdm = self.soap_client.content.virtualDiskManager
            task = vdm.DeleteVirtualDisk(
                name=backend_disk.backing.fileName,
                datacenter=self.get_disk_datacenter(backend_disk),
            )
            try:
                pyVim.task.WaitForTask(task)
            except Exception:
                logger.exception('Unable to delete VMware disk. Disk ID: %s.', disk.id)
                raise VMwareBackendError('Unknown error.')
            signals.vm_updated.send(self.__class__, vm=disk.vm)

    def extend_disk(self, disk):
        """
        Increase disk capacity.

        :param disk: Virtual disk to be extended.
        :type disk: :class:`waldur_vmware.models.Disk`
        """
        backend_vm = self.get_backend_vm(disk.vm)
        backend_disk = self.get_backend_disk(disk)

        virtual_disk_spec = vim.vm.device.VirtualDeviceSpec()
        virtual_disk_spec.operation = vim.vm.device.VirtualDeviceSpec.Operation.edit
        virtual_disk_spec.device = backend_disk
        virtual_disk_spec.device.capacityInKB = disk.size * 1024
        virtual_disk_spec.device.capacityInBytes = disk.size * 1024 * 1024

        spec = vim.vm.ConfigSpec()
        spec.deviceChange = [virtual_disk_spec]
        task = backend_vm.ReconfigVM_Task(spec=spec)

        try:
            pyVim.task.WaitForTask(task)
        except Exception:
            logger.exception('Unable to extend VMware disk. Disk ID: %s.', disk.id)
            raise VMwareBackendError('Unknown error.')
        signals.vm_updated.send(self.__class__, vm=disk.vm)

    def get_object(self, vim_type, vim_id):
        """
        Get object by type and ID from SOAP client.
        """
        content = self.soap_client.content
        try:
            items = [item for item in content.viewManager.CreateContainerView(
                content.rootFolder, [vim_type], recursive=True
            ).view]
        except Exception:
            logger.exception('Unable to get VMware object. Type: %s, ID: %s.', vim_type, vim_id)
            raise VMwareBackendError('Unknown error.')
        for item in items:
            if item._moId == vim_id:
                return item

    def get_backend_vm(self, vm):
        """
        Get virtual machine object from SOAP client.

        :param vm: Virtual machine.
        :type vm: :class:`waldur_vmware.models.VirtualMachine`
        :rtype: :class:`pyVmomi.VmomiSupport.vim.VirtualMachine`
        """
        return self._get_backend_vm(vm.backend_id)

    def get_vm_tools_state(self, backend_id):
        """
        Get running status of VMware Tools.

        :param backend_id: Virtual machine identifier.
        :type backend_id: str
        :rtype: str
        """
        backend_vm = self._get_backend_vm(backend_id)
        backend_tools_state = backend_vm.guest.toolsRunningStatus
        if backend_tools_state == 'guestToolsExecutingScripts':
            return models.VirtualMachine.ToolsStates.STARTING
        elif backend_tools_state == 'guestToolsNotRunning':
            return models.VirtualMachine.ToolsStates.NOT_RUNNING
        elif backend_tools_state == 'guestToolsRunning':
            return models.VirtualMachine.ToolsStates.RUNNING

    def get_vm_tools_installed(self, backend_id):
        """
        Check if VMware Tools are installed.

        :param backend_id: Virtual machine identifier.
        :type backend_id: str
        :rtype: bool
        """
        backend_vm = self._get_backend_vm(backend_id)
        return backend_vm.config.tools.toolsInstallType != 'guestToolsTypeUnknown'

    def _get_backend_vm(self, backend_id):
        return self.get_object(vim.VirtualMachine, backend_id)

    def get_backend_disk(self, disk):
        """
        Get virtual disk object from SOAP client.

        :param disk: Virtual disk.
        :type disk: :class:`waldur_vmware.models.Disk`
        :rtype: :class:`pyVmomi.VmomiSupport.vim.vm.device.VirtualDisk`
        """
        backend_vm = self.get_backend_vm(disk.vm)
        for device in backend_vm.config.hardware.device:
            if isinstance(device, vim.VirtualDisk) and str(device.key) == disk.backend_id:
                return device

    def get_disk_datacenter(self, backend_disk):
        """
        Find the datacenter where virtual disk is located.

        :param backend_disk: Virtual disk object returned by SOAP API.
        :type backend_disk: :class:`pyVmomi.VmomiSupport.vim.vm.device.VirtualDisk`
        :return: VMware datacenter where disk is located.
        :rtype: :class:`pyVmomi.VmomiSupport.vim.Datacenter`
        """
        parent = backend_disk.backing.datastore.parent
        while parent and not isinstance(parent, vim.Datacenter):
            parent = parent.parent
        return parent

    @log_backend_action()
    def pull_disk(self, disk, update_fields=None):
        """
        Pull virtual disk from REST API and update its information in local database.

        :param disk: Virtual disk database object.
        :type disk: :class:`waldur_vmware.models.Disk`
        :param update_fields: iterable of fields to be updated
        :return: None
        """
        import_time = timezone.now()
        imported_disk = self.import_disk(disk.vm.backend_id, disk.backend_id, save=False)

        disk.refresh_from_db()
        if disk.modified < import_time:
            if not update_fields:
                update_fields = models.Disk.get_backend_fields()

            update_pulled_fields(disk, imported_disk, update_fields)

    def import_disk(self, backend_vm_id, backend_disk_id, save=True, service_project_link=None):
        """
        Import virtual disk by its ID.

        :param backend_vm_id: Virtual machine identifier
        :type backend_vm_id: str
        :param backend_disk_id: Virtual disk identifier
        :type backend_disk_id: str
        :param save: Save object in the database
        :type save: bool
        :param service_project_link: Service project link model object
        :rtype: :class:`waldur_vmware.models.Disk`
        """
        try:
            backend_disk = self.client.get_disk(backend_vm_id, backend_disk_id)
        except VMwareError as e:
            reraise(e)
            return

        disk = self._backend_disk_to_disk(backend_disk, backend_disk_id)
        if service_project_link is not None:
            disk.service_project_link = service_project_link
        if save:
            disk.save()

        return disk

    def _backend_disk_to_disk(self, backend_disk, backend_disk_id):
        """
        Build database model object for virtual disk from REST API spec.

        :param backend_disk: virtual disk specification
        :type backend_disk: dict
        :param backend_disk_id: Virtual disk identifier
        :type backend_disk_id: str
        :rtype: :class:`waldur_vmware.models.Disk`
        """
        return models.Disk(
            backend_id=backend_disk_id,
            name=backend_disk['label'],
            # Convert disk size from bytes to MiB
            size=backend_disk['capacity'] / 1024 / 1024,
            state=models.Disk.States.OK,
        )

    def get_console_url(self, vm):
        """
        Generates a virtual machine's remote console URL (VMRC)

        :param vm: Virtual machine.
        :type vm: :class:`waldur_vmware.models.VirtualMachine`
        """
        ticket = self.soap_client.content.sessionManager.AcquireCloneTicket()
        return 'vmrc://clone:{ticket}@{host}/?moid={vm}'.format(
            ticket=ticket, host=self.host, vm=vm.backend_id)

    def get_web_console_url(self, vm):
        """
        Generates a virtual machine's web console URL (WMKS)

        :param vm: Virtual machine.
        :type vm: :class:`waldur_vmware.models.VirtualMachine`
        """
        backend_vm = self.get_backend_vm(vm)
        ticket = backend_vm.AcquireMksTicket()
        params = {
            'host': ticket.host,
            'port': ticket.port,
            'ticket': ticket.ticket,
            'cfgFile': ticket.cfgFile,
            'thumbprint': ticket.sslThumbprint,
            'vmId': vm.backend_id,
            'encoding': 'UTF-8'
        }
        return 'wss://{host}/ui/webconsole/authd?{params}'.format(
            host=ticket.host,
            params=urlparse.urlencode(params)
        )



from django.apps import AppConfig


class MarketplaceRancherConfig(AppConfig):
    name = 'waldur_mastermind.marketplace_rancher'
    verbose_name = 'Marketplace Rancher'

    def ready(self):
        from waldur_mastermind.marketplace.plugins import manager

        from . import PLUGIN_NAME, processors

        manager.register(offering_type=PLUGIN_NAME,
                         create_resource_processor=processors.RancherCreateProcessor,
                         delete_resource_processor=processors.RancherDeleteProcessor)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.14 on 2018-08-16 10:18
from __future__ import unicode_literals

import django.contrib.postgres.fields.jsonb
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import model_utils.fields
import waldur_core.core.fields
import waldur_core.core.validators


class Migration(migrations.Migration):

    dependencies = [
        ('support', '0021_protect_issue_user'),
    ]

    operations = [
        migrations.CreateModel(
            name='OfferingTemplate',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('name', models.CharField(max_length=150, unique=True, verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('config', django.contrib.postgres.fields.jsonb.JSONField()),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.AddField(
            model_name='offering',
            name='template',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='support.OfferingTemplate'),
        ),
        migrations.AlterField(
            model_name='offering',
            name='type',
            field=models.CharField(editable=False, max_length=255),
        ),
    ]



from django.conf import settings
from django.core.cache import cache

QUOTA_NAME = 'freeipa_quota'

CACHE_KEY = 'waldur_freeipa_syncing_groups'


def is_syncing():
    """
    This function checks if task is already running.
    """
    return cache.get(CACHE_KEY)


def renew_task_status():
    """
    This function sets lock with timeout. Lock is valid only for 1 minute.
    Then it should be renewed. Otherwise, lock is released.
    """
    cache.set(CACHE_KEY, True, 60)


def release_task_status():
    cache.set(CACHE_KEY, False)


def get_names(full_name):
    full_name_list = full_name.split()
    initials = ''

    if full_name_list:
        first_name = full_name_list[0]
        initials = first_name[0]
    else:
        first_name = 'N/A'

    if len(full_name_list) > 1:
        last_name = full_name_list[-1]
        initials += last_name[0]
    else:
        last_name = 'N/A'
    return first_name, last_name, initials


def generate_username(username):
    # Prepend username suffix
    prefix = settings.WALDUR_FREEIPA['USERNAME_PREFIX']
    if prefix:
        username = '%s%s' % (prefix, username)
    return username



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2018-01-24 12:59
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('support', '0012_attachment'),
    ]

    operations = [
        migrations.AddField(
            model_name='attachment',
            name='author',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='attachments', to='support.SupportUser'),
        ),
        migrations.AddField(
            model_name='attachment',
            name='file_size',
            field=models.PositiveIntegerField(blank=True, null=True, verbose_name='Filesize, B'),
        ),
        migrations.AddField(
            model_name='attachment',
            name='mime_type',
            field=models.CharField(blank=True, max_length=100, verbose_name='MIME type'),
        ),
        migrations.AddField(
            model_name='attachment',
            name='thumbnail',
            field=models.FileField(blank=True, null=True, upload_to='support_attachments_thumbnails'),
        ),
    ]



import copy

from django.conf import settings
from django.test import override_settings

from . import models


def override_paypal_settings(**kwargs):
    plugin_settings = copy.deepcopy(settings.WALDUR_PAYPAL)
    plugin_settings.update(kwargs)
    return override_settings(WALDUR_PAYPAL=plugin_settings)


def convert_unit_of_measure(unit):
    if unit == 'quantity':
        return models.InvoiceItem.UnitsOfMeasure.QUANTITY
    else:
        return models.InvoiceItem.UnitsOfMeasure.AMOUNT


def get_invoice_item_quantity(item):
    if item.unit == 'quantity':
        return item.quantity
    else:
        return item.usage_days



from django.utils.translation import gettext_lazy as _
from rest_framework import decorators, viewsets, serializers as rf_serializers, response, status

from waldur_core.core import exceptions as core_exceptions, validators as core_validators
from waldur_core.structure import views as structure_views

from . import filters, models, serializers, executors


class AmazonServiceViewSet(structure_views.BaseServiceViewSet):
    queryset = models.AWSService.objects.all()
    serializer_class = serializers.ServiceSerializer
    import_serializer_class = serializers.InstanceImportSerializer

    def get_import_context(self):
        return {'resource_type': self.request.query_params.get('resource_type')}

    def get_serializer_class(self):
        from waldur_core.structure import SupportedServices

        if self.request.method == 'POST':
            resource_type = self.request.data.get('type')
            if resource_type == SupportedServices.get_name_for_model(models.Instance):
                return serializers.InstanceImportSerializer
            elif resource_type == SupportedServices.get_name_for_model(models.Volume):
                return serializers.VolumeImportSerializer
        return super(AmazonServiceViewSet, self).get_serializer_class()


class AmazonServiceProjectLinkViewSet(structure_views.BaseServiceProjectLinkViewSet):
    queryset = models.AWSServiceProjectLink.objects.all()
    serializer_class = serializers.ServiceProjectLinkSerializer


class RegionViewSet(structure_views.BaseServicePropertyViewSet):
    queryset = models.Region.objects.all()
    serializer_class = serializers.RegionSerializer
    filter_class = filters.RegionFilter
    lookup_field = 'uuid'


class ImageViewSet(structure_views.BaseServicePropertyViewSet):
    queryset = models.Image.objects.all()
    serializer_class = serializers.ImageSerializer
    filter_class = filters.ImageFilter
    lookup_field = 'uuid'


class SizeViewSet(viewsets.ReadOnlyModelViewSet):
    queryset = models.Size.objects.all()
    serializer_class = serializers.SizeSerializer
    filter_class = filters.SizeFilter
    lookup_field = 'uuid'


class InstanceViewSet(structure_views.ResourceViewSet):
    queryset = models.Instance.objects.all()
    filter_class = filters.InstanceFilter
    serializer_class = serializers.InstanceSerializer
    create_executor = executors.InstanceCreateExecutor

    delete_executor = executors.InstanceDeleteExecutor
    destroy_validators = [core_validators.StateValidator(models.Instance.States.OK, models.Instance.States.ERRED)]

    def perform_create(self, serializer):
        instance = serializer.save()
        volume = instance.volume_set.first()

        self.create_executor.execute(
            instance,
            image=serializer.validated_data.get('image'),
            size=serializer.validated_data.get('size'),
            ssh_key=serializer.validated_data.get('ssh_public_key'),
            volume=volume
        )

    @decorators.detail_route(methods=['post'])
    def start(self, request, uuid=None):
        instance = self.get_object()
        executors.InstanceStartExecutor().execute(instance)
        return response.Response({'status': _('start was scheduled')}, status=status.HTTP_202_ACCEPTED)

    start_validators = [core_validators.StateValidator(models.Instance.States.OK),
                        core_validators.RuntimeStateValidator('stopped')]
    start_serializer_class = rf_serializers.Serializer

    @decorators.detail_route(methods=['post'])
    def stop(self, request, uuid=None):
        instance = self.get_object()
        executors.InstanceStopExecutor().execute(instance)
        return response.Response({'status': _('stop was scheduled')}, status=status.HTTP_202_ACCEPTED)

    stop_validators = [core_validators.StateValidator(models.Instance.States.OK),
                       core_validators.RuntimeStateValidator('running')]
    stop_serializer_class = rf_serializers.Serializer

    @decorators.detail_route(methods=['post'])
    def restart(self, request, uuid=None):
        instance = self.get_object()
        executors.InstanceRestartExecutor().execute(instance)
        return response.Response({'status': _('restart was scheduled')}, status=status.HTTP_202_ACCEPTED)

    restart_validators = [core_validators.StateValidator(models.Instance.States.OK),
                          core_validators.RuntimeStateValidator('running')]
    restart_serializer_class = rf_serializers.Serializer

    @decorators.detail_route(methods=['post'])
    def resize(self, request, uuid=None):
        instance = self.get_object()
        serializer = self.get_serializer(instance, data=request.data)
        serializer.is_valid(raise_exception=True)
        serializer.save()

        new_size = serializer.validated_data.get('size')
        executors.InstanceResizeExecutor().execute(instance, size=new_size)
        return response.Response({'status': _('resize was scheduled')}, status=status.HTTP_202_ACCEPTED)

    resize_validators = [core_validators.StateValidator(models.Instance.States.OK)]
    resize_serializer_class = serializers.InstanceResizeSerializer


class VolumeViewSet(structure_views.ResourceViewSet):
    queryset = models.Volume.objects.all()
    serializer_class = serializers.VolumeSerializer
    create_executor = executors.VolumeCreateExecutor
    delete_executor = executors.VolumeDeleteExecutor

    def _has_instance(volume):
        if not volume.instance:
            raise core_exceptions.IncorrectStateException(_('Volume is already detached.'))

    @decorators.detail_route(methods=['post'])
    def detach(self, request, uuid=None):
        executors.VolumeDetachExecutor.execute(self.get_object())

    detach_validators = [core_validators.StateValidator(models.Volume.States.OK), _has_instance]
    detach_serializer_class = rf_serializers.Serializer

    @decorators.detail_route(methods=['post'])
    def attach(self, request, volume, uuid=None):
        serializer = self.get_serializer(volume, data=request.data)
        serializer.is_valid(raise_exception=True)
        serializer.save()

        executors.VolumeAttachExecutor.execute(volume)

    attach_validators = [core_validators.StateValidator(models.Volume.States.OK)]
    attach_serializer_class = serializers.VolumeAttachSerializer
